/*****************************************************************************/
/*
                                  DCL.c

*************
** CAUTION **
*************

THIS MODULE IS TASK-ORIENTED, NOT REQUEST-ORIENTED.

That is, most of the functions take a pointer to DCL task rather than a
pointer to request as do other modules. The reason is simple. In this module
the more-permanent data-structures are those associated with the DCL
subprocesses, which persist for multiple requests. With the requests transient
the DCL subprocesses must be managed in their absence. Hence requests are
associated with DCL subprocesses not vice-versa.


OVERVIEW
--------
Provides multi-threaded, concurrent HTTPd subprocesses executing DCL.

This module never returns a valid status and ALWAYS calls the supplied next
task (AST) function.  This should check for a generated error message to
determine is there were any problems.

The DCL can either be in the form of a command, or a procedure or executable
image file specification. Both should not be supplied, but file specifications
have precedence. If a file specfication is supplied the module verifies its
existance, and if not qualified with an extension, looks for a procedure first
(".COM"), then an executable image (".EXE"), then through any user-defined
list of file types (extensions) and associated scripting executables (e.g.
".PL" associated with the verb "PERL"). Furthermore, the DCL can be executed
either standalone or as a CGI script (indicated by the presence of a script
name in the request data structure). If a CGI script, then the output stream
is parsed for header information, if not a script then the stream is just
checked for correct carriage control (terminated by a newline). CGI variables
are created for both standalone DCL and scripts, although some (e.g.
WWW_SCRIPT_NAME) will be empty, and meaningless, for standalone DCL. 

The AST-driven nature makes the code a little more difficult to follow, but 
creates a powerful, event-driven, multi-threaded server.  All of the 
necessary functions implementing this module are designed to be non-blocking. 

All of these functions are designed to be, and should be, called at AST 
delivery level, which means they cannot be interrupted at the same level (in 
this case USER mode), and so their actions are essentially atomic at that 
level, meaning no additional synchronization is required for such activities 
as thread disposal, etc.

The HTTPd can maintain a number of subprocesses limited only by its process 
quotas, memory is dynamically allocated and there are no fixed data structures 
related to subprocess management. 

The use of byte-streams (effectively "pipes") allows even DCL procedures to
output as HTTP servers, without the need for explicit network I/O. 

Four mailboxes are created for each subprocess' IPC:

  1.  A mailbox connected to its SYS$COMMAND.  This is used to pass DCL
      commands and/or other data to the subprocess.  It effectively allows
      the HTTPd to control the activities of the subprocess this way.

  2.  A mailbox connected to its SYS$OUTPUT.  This recieves records from
      the subprocess, if required appends the HTTP-required carriage-control
      (single <LF>), then sends the record to the client via the network.
      This allows even DCL procedures to supply a correct output stream to
      the client (see next paragraph).

      If the first line from a script is an HTTP status line (e.g.
      "HTTP/1.0 200 ok") then HTTPD assumes the script will be supplying
      a complete HTTP data stream, including full header and required
      carriage control (single <LF> terminating each line).  If the first
      line is not a HTTP status line it assumes CGI script output compliance
      and also ensures each record (line) received has correct HTTP
      carriage-control.

      This stream also attempts to maintain compliance with CGI scripting.
      If the first line output by a script is not an HTTP status line it
      creates and sends one to the client before sending the first line.

  3.  A mailbox defined for the subprocess by the name HTTP$INPUT.
      This may be used to read the request data steam sent by the client.
      Note that from v4.2 this is also the SYS$INPUT <stdin> stream as well,
      which appears to be the more common CGI implementation. As of version
      4.3 the default behaviour is to supply only the request body to the
      script (CGI standard and the more common implmentation).  Using the 
      configuration parameter [DclFullRequest] it is possible to revert
      to the previous behaviour of supplying the header then the body.
      (It's an easy enough modification for most scripts to skip the header by
      reading until the header-separating empty (blank) line is encountered.)

  4.  A mailbox defined by the name CGIPLUSIN.  This allows CGIplus scripts
      to read a stream of CGI variable information.  Each request begins with
      a comment line "!", which can be used for request start synchronisation
      and may always be discarded, and ends with an empty record (blank line),
      with a variable number of records in between.

The script subprocesses can use the basic CGI variables (VMS CERN-like) and 
behave very much like a CGI script.

That is, if a script wants to be CGI-compliant it provides as the first line a 
"Status:", "Content-type:" or a "Location:" then a blank line.   If the first 
line output by a script is a "Content-Type:" header line an HTTP "200" status 
line is prepended.  If the first line output by a script is a "Location:" 
redirection header line the redirection is processed to ensure CERN HTTPD/CGI 
behaviour.  An HTTP "302" status line is prepended if not a local redirection. 
If none of these, HTTPD creates a complete HTTP header comprising status line, 
"Content-Type: text/plain" and blank line (this is an extension of CERN HTTPD 
behaviour). 

If the first characters are "HTTP/1.0 ..." the script will be considered to be 
supplying the raw HTTP stream and record boundaries, carriage-control, etc., 
are of no further concern to the module.  This is the equivalent of a
"no-parse-header" script.  If CGI-compliant each record should represent a 
line of output.   That is lines should not be buffered together and sent as a 
block unless the script is supplying a raw HTTP data stream. 


CGI VARIABLES
-------------
See CGI.c module.


CGI-PLUS
--------
CGI plus lower latency, plus greater efficiency, plus less system impact!

CGIplus attempts to eliminate the overhead associated with creating the
subprocess and then executing the image of a CGI script.  It does this by
allowing the subprocess and associated image to continue executing in-between
uses, eliminating the startup overhead.  This both reduces the load on the
system and the request latency where one of these scripts is involved.
In this sense these advantages parallel those offered by commercial HTTP
server-integration APIs, such as Netscape NSAPI and Microsoft ISAPI.

The script interface is still largely CGI, which means a new API does not need
to be learned and that existing CGI scripts are simple to modify. The script
must read the CGI variables from CGIPLUSIN. They are supplied as a series of
records (lines) containing the CGI variable name, an equate symbol and then
the variable value. This format may be easily parsed and as the value contains
no encoded characters may be directly used. An empty record (blank line)
indicates the end of the request information. The request may be processed at
that stage.

After processing the CGIplus script can loop, waiting to read the details of
the next request from CGIPLUSIN. The first line read may ALWAYS be discarded
allowing a read to be used as a simple synchronization mechanism.

HTTP output (to the client) is written to SYS$OUTPUT (stdout). End of output
MUST be indicated by writing a special EOF string to the output stream. This
is a KLUDGE, and the least elegant part of CGIplus design, BUT it is also the
simplest implementation and should work in all but the most exceptional
circumstances. The special EOF string has been chosen to be most unlikely to
occur in normal output (a hopefully unique 248 bit sequence), but there is
still a very, very small chance! The CGIplus EOF string is obtained from the
logical name CGIPLUSEOF, defined in the subprocess's process table, using the
language's equivalent of F$TRNLNM(), SYS$TRNLNM(), or a getenv() call in the C
Language. This string will always contain less than 64 characters and comprise
only common, printable characters. It must be written at the conclusion of a
request's output to the output stream as a single record (line) but may also
contain a <CR><LF> or just <LF> trailing carriage-control (to allow for
programming language constraints). See examples in HT_ROOT:[SRC.CGIPLUS]

HTTP input (raw request stream, header and any body) is still available to the
CGIplus script.

Multiple CGIplus scripts may be executing in subprocesses at any one time.
This includes multiple instances of any particular script.  It is the server's
task to track these, distributing appropriate requests to idle subprocesses,
monitoring those currently processing requests, creating new instances if and
when necessary, and deleting the least-used, idle CGIplus subprocesses when
configurable thresholds are reached.

A CGIplus script can be terminated by the server at any time (the subprocess
SYS$DELPRC()ed) so resources should be largely quiescent when not actually
processing. CGIplus subprocesses may also be terminated from the command line
using STOP/ID=. The server administration menu provides a simple mechansim to
purge (stop) all CGIplus processes, allowing the server to be flushed of all
subprocesses. This can be useful if some new compilation of a CGIplus script
executable needs to made available.

CGIplus scripts are differentiated from "normal" CGI scripts in the mapping
rule configuration file using the "script+" and "exec+" directives.  Of course
it would be possible to design a script to simply modify it's behaviour so it
was possible to execute in both environments.  Simply detecting the presence
or absence of one of the "normal" CGI variables (i.e. DCL symbols,
e.g. WWW_PATH_INFO, WWW_CGI_GATEWAY_INTERFACE, etc.) would be sufficient
indication.

See examples in HT_ROOT:[SRC.CGIPLUS]

April 1998 Note:  It has been observed that under rare circumstances a
persistant subprocess script-serviced request can die unexpectedly.  This has
been isolated to the following scenario.  A request underway results in a
CGIplus script terminating and the subprocess exiting.  Any output in the
script's C RTL buffers is flushed during exit, including possibly the CGI
output EOF indicator, generating an I/O AST which is serviced indicating the
request is complete and the CGIplus script is ready for reuse.  In fact it
isn't because the process is running-down.  Before the subprocess termination
AST can be processed another AST containing a request for that same CGIplus
script is serviced.  The DCL task structure is allocated to the new request but
shortly, possibly immediately after, that DCL task receives the termination AST
and is run-down.  The request receives no output and Netscape Navigator for
instance reports "Document contains no data".  This has been addressed by
checking whether the script has begun processing (by reading from the COMMAND
or CGIPLUSIN variable stream.  If this stream has not been read from it is
considered the above has happened and the script request is resubmitted to
DclBegin().  A limit is placed on the number of times this may happen in
succession, to prevent an errant script from producing a runaway condition in
the server.


CGIPLUS-enabled CALLOUT
-----------------------
The CGIplus input/output streams are used to provide support for the subprocess
to "escape" from the normal SYS$OUTPUT stream and send records directly to a
server function for interpretation and processing.  In turn this function can
return records to the script via the CGIPLUSIN stream.  This works for *BOTH*
CGIplus *AND* vanilla CGI scripting!

The script indicates it wishes to suspend normal output by providing a record
containing the string found in the subprocess' CGIPLUSESC logical.  End of this
mode of communication is indicated by a record containing the string from the
CGIPLUSEOT logical (after which output to the client is resumed).  These
sequences are generated and used in much the same way as CGIPLUSEOF.

This functionality is used to supports CGIplus scripts that act not in the
traditional role as a CGI script but as "agents" to perform some required
function transparently to request processing.  An obvious example is an
external authentication/authorization processor.  These agents execute in the
normal CGI environment, with the request's CGI variables available (with some
minor differences depending on authorization state), <stdout> output stream,
etc.  There are some behavioural constraints on agents but this general
approach confers the considerable benefits of being able to write and operate
these agents as if CGI/CGIplus scripts, even as (CG)ISAPI DLLs.

A default callout for a CGIplus script is provided.  This function provides a
number of operations that may be useful to special-purpose scripts.  The
responses provided by the server are always an HTTP-like code, 200 for success,
400 for script request error, etc., with trailing plain-text explanation.

CGIplus "callout" requests (case-insensitive):

  AUTH-FILE: string      authorize/deny SYSPRV access to the file specified
  LIFETIME: integer      set the CGIplus process lifetime, in minutes
                         (zero makes it immune to supervisor purging)
  MAP-FILE: string       map the supplied file to a path
  MAP-PATH: string       map the supplied path to a file
  NOOP:                  do nothing, just return a success status


ZOMBIES
-------
The reuse of DCL subprocesses for CGI scripting provides very significant
performance gains with very little _real_ possibility of undesirable
interaction between uses (where scripts are "well-behaved", which should be
all enviornments). When a non-zero zombie lifetime is specified DCL
subprocesses implicitly persist between uses for standard CGI and DCL (SSI)
commands as well as explicitly with CGIplus scripts. When not being used to
process a request these subprocesses are termed "zombies" ;^) they are neither
"alive" (executing a script and processing a request) nor are they "dead"
(subprocess deleted and task structure free). Instead the subprocess is in a
LEF state waiting for more input via CGIPLUSIN. A great deal of care is
taken to ensure there is no interaction between uses (all symbols are deleted,
output mailbox is emptied), but there can be no "iron-clad" guarantee. Use of
zombies (persistant DCL processes) is disabled by setting the appropriate
configuration parameter to zero. The same DCL EOF mechansism is used to signal
end-of-output in all process-persistant environments.


SCRIPT RUNTIME ENVIRONMENTS
---------------------------
File types (extensions) and associated scripting languages can be defined in
the configuration file.  The syntax is "type foreign-verb".  For example:

  [DclRunTime]
  .PL  $PERL_EXE:PERL.EXE
  .CGI PERL

Two are predefined, ".COM" for DCL procedures, and ".EXE" for executables.


LOGICAL NAMES
-------------
HTTPD$LOGIN     if defined, provides a command procedure executed immediately
                            before the script command/procedure/image
HTTPD$VERIFY    if defined, turns on script/DCL-level verify


VERSION HISTORY
---------------
22-JUN-2000  MGD  bugfix; HEAD requests specifying content-length
17-JUN-2000  MGD  fixed potential problem when setting 'DclSysOutputSize'
27-MAY-2000  MGD  add BYTLM check before creating mailboxes
08-APR-2000  MGD  some (VMS) Apache compatibility,
                  if(!cfptr->cfScript.Enabled)
04-MAR-2000  MGD  use NetWriteFaol(), et.al.
13-JAN-2000  MGD  add OPCOM messages
03-JAN-2000  MGD  support ODS-2 and ODS-5 using ODS module for script find
28-NOV-1999  MGD  provide HTTPD$LOGIN
10-OCT-1999  MGD  make SYS$COMMAND and CGIPLUSIN mailbox sizes configurable,
                  bugfix; DclSysOutputToClientAst()
28-AUG-1999  MGD  callout/agent support,
                  support Purveyor/Cern WWW_IN: and WWW_OUT:,
                  bugfix; sizeof(StopId)-1
12-JUN-1999  MGD  change some WatchData() to WatchDataDump()
12-FEB-1999  MGD  refine WATCH information
07-NOV-1998  MGD  WATCH facility
17-OCT-1998  MGD  script name cache,
                  error report support
19-SEP-1998  MGD  improve granularity of script search,
                  do not search for established CGIplus scripts
                  (this reduces CGIplus script activation time by 50%!!)
15-AUG-1998  MGD  replace per-subprocess timers with DclSupervisor()
27-MAY-1998  MGD  generate CGI variables only once for any one request
02-APR-1998  MGD  see "April 1998 Note" above,
                  report status 500/501 if script returns no output
28-MAR-1998  MGD  ensure script output is null-terminated (for CGI.C)
16-DEC-1997  MGD  generalized CGI processing unbundled into CGI.c module
06-DEC-1997  MGD  resolving a suspected inconsistent AST delivery situation
                  by requiring all $QIO()s with an AST routine to ensure any
                  queueing errors, etc. are reported via the AST routine
19-OCT-1997  MGD  extensible script run-time environment,
                  HTTP_ACCEPT_CHARSET, HTTP_FORWARDED and HTTP_HOST variables,
                  change in behaviour: CGI "Content-Type:" response bodies now
                  only have carriage-control checked/adjusted if "text/..."
10-SEP-1997  MGD  add "!'F$VERIFY(0)" to start of DCL in case verify on account
09-AUG-1997  MGD  ACCEPT_LANGUAGE variable
01-AUG-1997  MGD  allow supplying request header AS WELL AS body, or only body,
                  added AUTH_REALM, AUTH_GROUP, HTTP_ACCEPT, and 
                  REQUEST_TIME_GMT/LOCAL CGI variables
01-JUN-1997  MGD  Persistant DCL processes, CGIplus, new for HTTPd v4.2,
                  DCL.C completely re-designed!
26-APR-1997  MGD  bugfix; serious flaw POST content handling since v4.0
                  to REQUEST.C, PUT.C in version 4.0 (rewrite of HTTP$INPUT)
01-FEB-1997  MGD  HTTPd version 4
14-NOV-1996  MGD  bugfix; no status was being returned after "DELETE/SYMBOL X"
                  in DclSysCommandStringSymbol()
06-APR-1996  MGD  miscellaneous refinements
26-MAR-1996  MGD  added WWW_HTTP_AUTHENTICATION, scripts can now authenticate
01-DEC-1995  MGD  HTTPd version 3
19-SEP-1995  MGD  changed carriage-control on records from <CR><LF> (the strict
                  HTTP requirement) to single newline (<LF>, de facto standard)
                  This will be slightly more efficient, and "more compliant"!
21-APR-1995  MGD  bugfix; DclSysOutputAst()
03-APR-1995  MGD  added remote user authentication and CGI symbol
20-MAR-1995  MGD  bugfix; DclQioHttpInput()
20-DEC-1994  MGD  initial development as a module for multi-threaded daemon
*/
/*****************************************************************************/

/* standard C header files */
#include <ctype.h>
#include <stdio.h>
#include <string.h>

/* VMS related header files */

/* cmbdef.h is not defined for VAXC 3.n */
#define CMB$M_READONLY 0x01
#define CMB$M_WRITEONLY 0x02

#include <descrip.h>
#include <dvidef.h>

#include <iodef.h>
/* these should be, but are not defined for VAXC 3.n */
#define IO$M_READERCHECK 0x100
#define IO$M_WRITERCHECK 0x200

#include <jpidef.h>
#include <prvdef.h>
#include <rms.h>

#include <ssdef.h>
/* these should be, but are not defined for VAXC 3.n */
#define SS$_NOREADER 9412
#define SS$_NOWRITER 9420

#include <stsdef.h>
#include <syidef.h>

/* application header files */
#include "wasd.h"

#define WASD_MODULE "DCL"

/* provides code to support vanilla-CGI use of CGIplus callouts */
#define CGIPLUS_CALLOUT_FOR_CGI 1

/* provides APACHE_INPUT: stream for VMS Apache compatibility */
#define STREAMS_FOR_APACHE 1

/* provides WWW_IN: and WWW_OUT: streams for Purveyor/Cern compatibility */
#define STREAMS_FOR_PURVEYOR_AND_CERN 1

/******************/
/* global storage */
/******************/

boolean  DclUseZombies;

int  DclCgiPlusLifeTimePurgeCount,
     DclCgiPlusInSize,
     DclSysCommandSize,
     DclSysOutputSize,
     DclHitHardLimitCount,
     DclMailboxBytLmRequired,
     DclPurgeCount,
     DclPurgeAllSubprocessesCount,
     DclPurgeScriptNameCacheCount,
     DclSoftLimitPurgeCount,
     DclSubprocessCount,
     DclSubprocessHardLimit,
     DclSubprocessSoftLimit,
     DclZombieLifeTimePurgeCount;

struct ListHeadStruct  DclTaskList;

struct ListHeadStruct  DclScriptNameCacheList;

#define DEFAULT_CGI_VARIABLE_PREFIX "WWW_"
char DclCgiVariablePrefix [32] = DEFAULT_CGI_VARIABLE_PREFIX;
int DclCgiVariablePrefixLength = sizeof(DEFAULT_CGI_VARIABLE_PREFIX)-1;

/********************/
/* external storage */
/********************/

#ifdef DBUG
extern boolean Debug;
#else
#define Debug 0 
#endif

extern boolean  OperateWithSysPrv;

extern int  WatchEnabled,
            HttpdBytLmAvailable,
            NetAcceptBytLmRequired,
            NetListenBytLmRequired,
            NetReadBufferSize,
            OpcomMessages,
            OutputBufferSize,
            ServerPort;

extern unsigned long  SysPrvMask[];

extern char  ErrorSanityCheck[],
             HttpProtocol[],
             ServerHostPort[],
             SoftwareID[],
             Utility[];

extern struct AccountingStruct Accounting;
extern struct ConfigStruct Config;
extern struct MsgStruct Msgs;

/*****************************************************************************/
/*
Set and ensure limits are reasonable at server startup.
*/ 

DclInit ()

{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DclInit()\n");

   /* command line overrides configuration, default fallbacks */
   if (!DclSysCommandSize) DclSysCommandSize = Config.cfBuffer.SizeDclCommand;
   if (!DclSysCommandSize) DclSysCommandSize = DEFAULT_DCL_SYSCOMMAND_SIZE;
   if (!DclSysOutputSize) DclSysOutputSize = Config.cfBuffer.SizeDclOutput;
   /* if not specified then make it the same as the network buffer */
   if (!DclSysOutputSize) DclSysOutputSize = OutputBufferSize;
   if (!DclSysOutputSize) DclSysOutputSize = DEFAULT_DCL_SYSOUTPUT_SIZE;
   if (!DclCgiPlusInSize) DclCgiPlusInSize = Config.cfBuffer.SizeDclCgiPlusIn;
   if (!DclCgiPlusInSize) DclCgiPlusInSize = DEFAULT_DCL_CGIPLUSIN_SIZE;
   /* HTTP$INPUT size is determined by network read buffer size! */

   DclSubprocessSoftLimit = Config.cfScript.SubprocessSoftLimit;
   if (DclSubprocessSoftLimit <= 0 || DclSubprocessSoftLimit > 99)
      DclSubprocessSoftLimit = Config.cfServer.BusyLimit;

   DclSubprocessHardLimit = Config.cfScript.SubprocessHardLimit;
   if (DclSubprocessHardLimit <= DclSubprocessSoftLimit ||
       DclSubprocessHardLimit > 99)
      DclSubprocessHardLimit = DclSubprocessSoftLimit +
                               (DclSubprocessSoftLimit / 4);

   if (Config.cfScript.ZombieLifeTime) DclUseZombies = true;
}

/*****************************************************************************/
/*
This function does not return a status value. If an error occurs the
'NextTaskFunction()' is executed. The calling routine may assume that this
module will always execute the 'NextTaskFunction()' at some stage. No need to
look for an established CGIplus script, we know it's there otherwise it never
would have been allocated. For new CGIplus scripts, as well as for standard
CGI scripts, first find the script file (as well as confirming it does exist!)
Already established CGIplus scripts and DCL commands can begin I/O with the
subprocess immediately.
*/ 

DclBegin
(
struct RequestStruct *rqptr,
void *NextTaskFunction,
char *DclCommand,
char *ScriptName,
char *CgiScriptFileName,
char *CgiPlusScriptFileName,
void *CgiPlusCalloutFunction
)
{
   register struct DclTaskStruct  *tkptr;

   int  TaskType;

   /*********/
   /* begin */
   /*********/

   if (Debug)
   {
      fprintf (stdout,
"DclBegin()\n\
DclCommand     |%s|\n\
ScriptName     |%s|\n\
CgiScriptFileName |%s|\n\
CgiPlusScriptFileName |%s|\n\
CalloutFunction: %d\n\
PathInfoPtr    |%s|\n\
QueryStringPtr |%s|\n",
      DclCommand, ScriptName, CgiScriptFileName, CgiPlusScriptFileName,
      CgiPlusCalloutFunction, rqptr->rqHeader.PathInfoPtr,
      rqptr->rqHeader.QueryStringPtr);
   }

   if (rqptr->rqResponse.ErrorReportPtr != NULL)
   {
      /* previous error, cause threaded processing to unravel */
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }

   if (!Config.cfScript.Enabled)
   {
      Accounting.RequestForbiddenCount++;
      rqptr->rqResponse.HttpStatus = 403;
      ErrorGeneral (rqptr, MsgFor(rqptr,MSG_GENERAL_DISABLED), FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }

   if (CgiScriptFileName != NULL && CgiScriptFileName[0])
      TaskType = DCL_TASK_TYPE_CGI_SCRIPT;
   else
   if (CgiPlusScriptFileName != NULL && CgiPlusScriptFileName[0])
      TaskType = DCL_TASK_TYPE_CGIPLUS_SCRIPT;
   else
   if (DclCommand != NULL && DclCommand[0])
      TaskType = DCL_TASK_TYPE_CLI;
   else
      ErrorExitVmsStatus (SS$_BUGCHECK, ErrorSanityCheck, FI_LI);

   if (rqptr->WatchItem &&
       (WatchEnabled & WATCH_RESPONSE))
   {
      switch (TaskType)
      {
         case DCL_TASK_TYPE_CGI_SCRIPT :
            WatchThis (rqptr, FI_LI, WATCH_RESPONSE,
                       "SCRIPT CGI !AZ !AZ",
                       ScriptName, CgiScriptFileName);
            break;
         case DCL_TASK_TYPE_CGIPLUS_SCRIPT :
            WatchThis (rqptr, FI_LI, WATCH_RESPONSE,
                       "SCRIPT CGIplus !AZ !AZ",
                       ScriptName, CgiPlusScriptFileName);
            break;
         case DCL_TASK_TYPE_CLI :
            WatchThis (rqptr, FI_LI, WATCH_RESPONSE,
                       "CLI !AZ", DclCommand);
            break;
         default :
               ErrorExitVmsStatus (SS$_BUGCHECK, ErrorSanityCheck, FI_LI);
      }
   }

   if (VMSnok (DclAllocateTask (rqptr, TaskType, ScriptName)))
   {
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }

   /* get a local pointer to the newly allocated DCL task structure */
   tkptr = rqptr->DclTaskPtr;

   tkptr->NextTaskFunction = NextTaskFunction;
   memcpy (&tkptr->LastUsedBinaryTime, &rqptr->rqTime.Vms64bit, 8);
   tkptr->WatchItem = rqptr->WatchItem;

   switch (tkptr->TaskType)
   {
      case DCL_TASK_TYPE_CGI_SCRIPT :

         if (CgiPlusCalloutFunction == NULL)
            tkptr->CgiPlusCalloutFunction = &DclCgiPlusDefaultCallout;
         else
            tkptr->CgiPlusCalloutFunction = CgiPlusCalloutFunction;

         strcpy (tkptr->ScriptName, ScriptName);
         strcpy (tkptr->ScriptFileName, CgiScriptFileName);
         tkptr->DclCommand[0] = '\0';

         /* reset CGI output processing */
         CgiOutput (rqptr, NULL, 0);

         if (!rqptr->AccountingDone)
         {
            rqptr->AccountingDone = true;
            Accounting.DoScriptCount++;
         }

         break;

      case DCL_TASK_TYPE_CGIPLUS_SCRIPT :

         if (CgiPlusCalloutFunction == NULL)
            tkptr->CgiPlusCalloutFunction = &DclCgiPlusDefaultCallout;
         else
            tkptr->CgiPlusCalloutFunction = CgiPlusCalloutFunction;

         /* reset CGI output processing */
         CgiOutput (rqptr, NULL, 0);

         if (!tkptr->ScriptName[0])
         {
            /* no need to reset these for established CGIplus scripts! */
            strcpy (tkptr->ScriptName, ScriptName);
            strcpy (tkptr->ScriptFileName, CgiPlusScriptFileName);
            tkptr->DclCommand[0] = '\0';

            if (!rqptr->AccountingDone)
            {
               rqptr->AccountingDone = true;
               Accounting.DoCgiPlusScriptCount++;
               if (tkptr->CgiPlusUsageCount > 1)
                  Accounting.DclCgiPlusReusedCount++;
            }
         }

         break;

      case DCL_TASK_TYPE_CLI :

         tkptr->CgiPlusCalloutFunction = NULL;
         strzcpy (tkptr->DclCommand, DclCommand, sizeof(tkptr->DclCommand));
         tkptr->ScriptName[0] = tkptr->ScriptFileName[0] = '\0';

         /* reset CGI output processing */
         CgiOutput (rqptr, NULL, 0);

         /* indicate this is *NOT* a script, but CLI executed DCL */
         rqptr->rqCgi.IsCliDcl =
            rqptr->rqCgi.RecordMode =
            rqptr->rqCgi.BufferRecords = true;

         if (!rqptr->AccountingDone)
         {
            rqptr->AccountingDone = true;
            Accounting.DoDclCommandCount++;
         }

         break;

      default :
            ErrorExitVmsStatus (SS$_BUGCHECK, ErrorSanityCheck, FI_LI);
   }


   if (tkptr->TaskType == DCL_TASK_TYPE_CGI_SCRIPT ||
       tkptr->TaskType == DCL_TASK_TYPE_CGIPLUS_SCRIPT)
   {
      if (!tkptr->SearchOds.ResFileName[0])
      {
         if (DclUseZombies) DclSearchScriptNameCache (tkptr, rqptr);
         /* if not found in the name cache then look in the file system! */
         if (!tkptr->SearchOds.ResFileName[0])
         {
            /* script is being requested, first look for the script file */
            tkptr->FindScriptState = DCL_FIND_SCRIPT_BEGIN;
            DclFindScript (tkptr);
            return;
         }
      }
   }
  
   /* must be an established CGIplus script or DCL command (from SSI module) */
   DclBeginScript (tkptr);
}

/*****************************************************************************/
/*
Allocate a DCL task structure to the request. All task structures are linked
together in a single list, function and state indicated by the various flags
and counters associated with each.

If a CGIplus task script is to be executed then check for an already
established, idle CGIplus task structure executing that particular script. If
none found, or CGIplus script not required, and zombies in use
(persistant-subprocesses) then look through the list for an idle zombie
subprocess. If no zombie available (or not enabled) then check if we have
reached the subprocess creation hard-limit. If not reached look through the
list for an existing but no-subprocess-executing DCL task structure. If none
found create an additional DCL task structure and add it to the list.

Initialize the task structure (no matter from which scan it originated) and if
necessary create a subprocess for it. (An obvious improvement to processing
would be to have multiple lists, but that will have to wait for another time
:^) If an error is encountered an error message is generated and the error
status returned. It is up to the calling routine to abort the processing.
*/

int DclAllocateTask
(
struct RequestStruct *rqptr,
int TaskType,
char* ScriptName
)
{
   register struct DclTaskStruct  *tkptr;
   register struct ListEntryStruct  *leptr;

   boolean  WatchThisOne;
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "DclAllocateTask() %d |%s|\n", TaskType, ScriptName);

   tkptr = NULL;

   if (rqptr->WatchItem && (WatchEnabled & WATCH_DCL))
      WatchThisOne = true;
   else
      WatchThisOne = false;

   if (TaskType == DCL_TASK_TYPE_CGIPLUS_SCRIPT)
   {
      /******************************************************/
      /* look for an unused instance of this CGIplus script */
      /******************************************************/

      for (leptr = DclTaskList.HeadPtr; leptr != NULL; leptr = leptr->NextPtr)
      {
         tkptr = (struct DclTaskStruct*)leptr;

         if (Debug) DclTaskItemDebug (leptr, tkptr);
 
         if (!tkptr->SubprocessPid ||
             tkptr->TaskType != DCL_TASK_TYPE_CGIPLUS_SCRIPT ||
             tkptr->QueuedSysCommand > tkptr->QueuedSysCommandAllowed ||
             tkptr->QueuedSysOutput ||
             tkptr->QueuedCgiPlusIn ||
             tkptr->QueuedHttpInput ||
             tkptr->RequestPtr ||
             tkptr->FindScript ||
             tkptr->MarkedForDelete ||
             !strsame (ScriptName, tkptr->ScriptName, -1))
         {
            tkptr = NULL;
            continue;
         }

         break;
      }

      if (Debug) fprintf (stdout, "CGIplus tkptr: %d\n", tkptr);

      if (tkptr != NULL && WatchThisOne)
         WatchThis (rqptr, FI_LI, WATCH_DCL,
                    "CGIPLUS idle pid !8XL", tkptr->SubprocessPid);
   }

   if (tkptr == NULL && DclUseZombies)
   {
      /************************/
      /* look for idle zombie */
      /************************/

      for (leptr = DclTaskList.HeadPtr; leptr != NULL; leptr = leptr->NextPtr)
      {
         tkptr = (struct DclTaskStruct*)leptr;

         if (Debug) DclTaskItemDebug (leptr, tkptr);

         if (!tkptr->SubprocessPid ||
             tkptr->TaskType == DCL_TASK_TYPE_CGIPLUS_SCRIPT ||
             tkptr->QueuedSysCommand ||
             tkptr->QueuedSysOutput ||
             tkptr->QueuedCgiPlusIn ||
             tkptr->QueuedHttpInput ||
             tkptr->FindScript ||
             tkptr->RequestPtr != NULL)
         {
            tkptr = NULL;
            continue;
         }
         tkptr->DclCommand[0] =
            tkptr->ScriptName[0] =
            tkptr->ScriptFileName[0] =
            tkptr->SearchOds.ResFileName[0] = '\0';
         break;
      }

      if (Debug) fprintf (stdout, "idle zombie tkptr: %d\n", tkptr);

      if (tkptr != NULL && WatchThisOne)
         WatchThis (rqptr, FI_LI, WATCH_DCL,
                    "ZOMBIE idle pid !8XL", tkptr->SubprocessPid);
   }

   if (tkptr == NULL && DclSubprocessCount >= DclSubprocessHardLimit)
   {
      if (WatchThisOne)
         WatchThis (rqptr, FI_LI, WATCH_DCL,
                    "SUBPROCESS hard-limit !UL", DclSubprocessCount);

      /* let's see if we can do something about it! */
      DclSubprocessPurge ();

      DclHitHardLimitCount++;
      ErrorGeneral (rqptr, MsgFor(rqptr,MSG_SCRIPT_HARD_LIMIT), FI_LI);
      return (STS$K_ERROR);
   }

   if (tkptr == NULL)
   {
      /********************************/
      /* look for free task structure */
      /********************************/

      for (leptr = DclTaskList.HeadPtr; leptr != NULL; leptr = leptr->NextPtr)
      {
         tkptr = (struct DclTaskStruct*)leptr;

         if (Debug) DclTaskItemDebug (leptr, tkptr);

         if (tkptr->SubprocessPid ||
             tkptr->QueuedSysCommand ||
             tkptr->QueuedSysOutput ||
             tkptr->QueuedCgiPlusIn ||
             tkptr->QueuedHttpInput ||
             tkptr->FindScript ||
             tkptr->RequestPtr != NULL)
         {
            tkptr = NULL;
            continue;
         }
         tkptr->DclCommand[0] =
            tkptr->ScriptName[0] =
            tkptr->ScriptFileName[0] =
            tkptr->SearchOds.ResFileName[0] = '\0';
         break;
      }

      if (Debug) fprintf (stdout, "free tkptr: %d\n", tkptr);
   }

   if (tkptr == NULL)
   {
      /* if we're getting short of subprocesses then start purging */
      if (DclSubprocessCount >= DclSubprocessSoftLimit)
      {
         if (WatchThisOne)
            WatchThis (rqptr, FI_LI, WATCH_DCL,
                       "SUBPROCESS soft-limit purge !UL", DclSubprocessCount);

         DclSubprocessPurge ();
      }
   }

   if (tkptr == NULL)
   {
      /*********************/
      /* create a new task */
      /*********************/

      tkptr = VmGet (sizeof(struct DclTaskStruct));
      if (Debug) fprintf (stdout, "tkptr: %d\n", tkptr);

      /*
         Allocate memory in the DCL task for SYS$OUTPUT buffer.
         Allow two bytes for carriage control and terminating null.
      */
      tkptr->SysOutputPtr = VmGet (DclSysOutputSize+3);
      tkptr->SysOutputSize = DclSysOutputSize;

      if (VMSnok (status = DclCreateMailboxes (tkptr)))
      {
         VmFree (tkptr->SysOutputPtr, FI_LI);
         VmFree (tkptr, FI_LI);
         rqptr->rqResponse.ErrorTextPtr = MsgFor(rqptr,MSG_SCRIPT_IPC);
         ErrorVmsStatus (rqptr, status, FI_LI);
         return (status);
      }

      ListAddTail (&DclTaskList, tkptr);
   }

   /*******************/
   /* initialize task */
   /*******************/

   /* associate the DCL task and the request */
   rqptr->DclTaskPtr = tkptr;
   tkptr->RequestPtr = rqptr;          

   tkptr->TaskType = TaskType;
   tkptr->TotalUsageCount++;
   tkptr->MarkedForDelete = false;

   switch (TaskType)
   {
      case DCL_TASK_TYPE_CGI_SCRIPT :

         if (Debug) fprintf (stdout, "CGI script\n");
         tkptr->SubprocessActivated = tkptr->SubprocessResponded = false;
         tkptr->CgiPlusUsageCount =
            tkptr->CgiEotLength =
            tkptr->CgiEscLength = 0;
         tkptr->CgiEot[0] = tkptr->CgiEsc[0] = '\0';

         /* limited life in the twilight zone */
         if (DclUseZombies && Config.cfScript.ZombieLifeTime)
         {
            /* plus one allows at least that period */
            tkptr->LifeTimeCount = Config.cfScript.ZombieLifeTime+1;
            DclSupervisor (false);
         }

         if (DclUseZombies)
         {
            /* always generate new strings for standard CGI scripts */
            CgiSequenceEof (tkptr->CgiEof, &tkptr->CgiEofLength);
            CgiSequenceEot (tkptr->CgiEot, &tkptr->CgiEotLength);
            CgiSequenceEsc (tkptr->CgiEsc, &tkptr->CgiEscLength);

            rqptr->rqCgi.EofPtr = tkptr->CgiEof;
            rqptr->rqCgi.EofLength = tkptr->CgiEofLength;
            rqptr->rqCgi.EotPtr = tkptr->CgiEot;
            rqptr->rqCgi.EotLength = tkptr->CgiEotLength;
            rqptr->rqCgi.EscPtr = tkptr->CgiEsc;
            rqptr->rqCgi.EscLength = tkptr->CgiEscLength;
         }

         break;

      case DCL_TASK_TYPE_CGIPLUS_SCRIPT :

         if (Debug) fprintf (stdout, "CGIplus script\n");
         tkptr->SubprocessActivated = tkptr->SubprocessResponded = false;
         tkptr->ZombieCount = 0;
         tkptr->CgiPlusUsageCount++;

         /* give it three-score years and ten if life-time is specified */
         if (Config.cfScript.CgiPlusLifeTime)
         {
            /* plus one allows at least that period */
            tkptr->LifeTimeCount = Config.cfScript.CgiPlusLifeTime+1;
            DclSupervisor (false);
         }

         /* CGIplus MUST retain original EOF/EOT/ESC until subprocess dies */
         if (!tkptr->CgiEofLength)
            CgiSequenceEof (tkptr->CgiEof, &tkptr->CgiEofLength);
         if (!tkptr->CgiEotLength)
            CgiSequenceEot (tkptr->CgiEot, &tkptr->CgiEotLength);
         if (!tkptr->CgiEscLength)
            CgiSequenceEsc (tkptr->CgiEsc, &tkptr->CgiEscLength);

         rqptr->rqCgi.EofPtr = tkptr->CgiEof;
         rqptr->rqCgi.EofLength = tkptr->CgiEofLength;
         rqptr->rqCgi.EotPtr = tkptr->CgiEot;
         rqptr->rqCgi.EotLength = tkptr->CgiEotLength;
         rqptr->rqCgi.EscPtr = tkptr->CgiEsc;
         rqptr->rqCgi.EscLength = tkptr->CgiEscLength;

         break;

      case DCL_TASK_TYPE_CLI :

         if (Debug) fprintf (stdout, "DCL command\n");
         tkptr->CgiPlusUsageCount =
            tkptr->CgiEotLength = tkptr->CgiEscLength = 0;
         tkptr->CgiEot[0] = tkptr->CgiEsc[0] = '\0';

         /* limited life in the twilight zone */
         if (DclUseZombies && Config.cfScript.ZombieLifeTime)
         {
            /* plus one allows at least that period */
            tkptr->LifeTimeCount = Config.cfScript.ZombieLifeTime+1;
            DclSupervisor (false);
         }

         if (DclUseZombies)
         {
            /* always generate a new EOF string for DCL commands */
            CgiSequenceEof (tkptr->CgiEof, &tkptr->CgiEofLength);
            rqptr->rqCgi.EofPtr = tkptr->CgiEof;
            rqptr->rqCgi.EofLength = tkptr->CgiEofLength;
         }

         break;

      default :
            ErrorExitVmsStatus (SS$_BUGCHECK, ErrorSanityCheck, FI_LI);
   }

   if (tkptr->SubprocessPid)
   {
      /*********************/
      /* subprocess exists */
      /*********************/

      return (SS$_NORMAL);
   }

   /*********************/
   /* create subprocess */
   /*********************/

   if (VMSok (status = DclCreateSubprocess (tkptr)))
   {
      if (WatchThisOne)
         WatchThis (rqptr, FI_LI, WATCH_DCL,
                    "SUBPROCESS create pid !8XL", tkptr->SubprocessPid);

      tkptr->ZombieCount = 0;
      DclSubprocessCount++;
      return (status);
   }
   else
   {
      if (WatchThisOne)
         WatchThis (rqptr, FI_LI, WATCH_DCL,
                    "SUBPROCESS create fail %X!8XL %!%M",
                    status, status);

      /* disassociate the DCL task and request structures */
      rqptr->DclTaskPtr = tkptr->RequestPtr = NULL;
      tkptr->WatchItem = 0;
      return (status);
   }
}

/*****************************************************************************/
/*
Search the script name cache for an entry with the same script name as the
task.  If not found just return with 'tkptr->SearchOds.ResFileName' still set
as an empty string.  If found and 'rqptr->PragmaNocache' is true then purge the
cache entry details by setting both script name and expanded file name to empty
strings and returning.  This effectively allows a request to purge a script
entry's details with the "reload" button at no great cost.  If the
'PragmaNoCache' is false the copy the expanded (original sys$search()ed for
file name) into the task structure.
*/ 

DclSearchScriptNameCache
(
struct DclTaskStruct *tkptr,
struct RequestStruct *rqptr
)
{
   register struct DclScriptNameCacheStruct  *captr;
   register struct ListEntryStruct  *leptr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "DclSearchScriptNameCache() |%s|\n", tkptr->ScriptName);

   for (leptr = DclScriptNameCacheList.HeadPtr;
        leptr != NULL;
        leptr = leptr->NextPtr)
   {
      captr = (struct DclScriptNameCacheStruct*)leptr;

      if (Debug)
         fprintf (stdout, "|%s|%s|%s|\n",
                  tkptr->ScriptName, captr->ScriptName,
                  captr->ResFileName);

      /* if request and cache script names not the same then next entry */
      if (!captr->ResFileName[0]) continue;
      if (!strsame(tkptr->ScriptName, captr->ScriptName, -1)) continue;

      /* found it */
      if (!rqptr->PragmaNoCache)
      {
         if (rqptr->WatchItem &&
             (WatchEnabled & WATCH_DCL))
            WatchThis (rqptr, FI_LI, WATCH_DCL,
                       "SEARCH name cache !AZ",
                       captr->ResFileName);

         /* copy the expanded (searched for) file name to the task */
         strcpy (tkptr->SearchOds.ResFileName, captr->ResFileName);
         tkptr->ScriptRunTimePtr = captr->ScriptRunTimePtr;
         captr->HitCount++;
         captr->UsageCount++;
         memcpy (&captr->LastBinaryTime, &rqptr->rqTime.Vms64bit, 8);
         return;
      }

      /* "reload" is used to purge the script name cache entry */
      captr->ResFileName[0] = '\0';
      captr->ScriptRunTimePtr = NULL;
      return;
   }

   if (rqptr->WatchItem &&
       (WatchEnabled & WATCH_DCL))
      WatchThis (rqptr, FI_LI, WATCH_DCL,
                 "SEARCH name cache NOT FOUND");
}

/*****************************************************************************/
/*
Add the name details of a script to the cache.  First check that no other
request has provided this same entry while this request was finding the script
details.  Then update an empty entry if it can be found in the list, or create
a new entry and add it to the list.
*/ 

DclUpdateScriptNameCache
(
struct DclTaskStruct *tkptr,
struct RequestStruct *rqptr
)
{
   register struct DclScriptNameCacheStruct  *captr;
   register struct ListEntryStruct  *leptr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "DclUpdateScriptNameCache() |%s|%s|\n",
               tkptr->ScriptName, tkptr->SearchOds.ResFileName);

   for (leptr = DclScriptNameCacheList.HeadPtr;
        leptr != NULL;
        leptr = leptr->NextPtr)
   {
      captr = (struct DclScriptNameCacheStruct*)leptr;

      if (Debug)
         fprintf (stdout, "|%s|%s|%s|\n",
                  tkptr->ScriptName, captr->ScriptName,
                  captr->ResFileName);

      if (!strsame (tkptr->ScriptName, captr->ScriptName, -1)) continue;

      /* if some other request has beaten us to it then just return! */
      if (captr->ResFileName[0]) return;
      /* this purged entry will do nicely */
      break;
   }

   if (leptr == NULL)
   {
      /* didn't find an instance of the script, look for an empty entry */
      captr = VmGet (sizeof(struct DclScriptNameCacheStruct));
      if (Debug) fprintf (stdout, "captr: %d\n", captr);
      ListAddTail (&DclScriptNameCacheList, captr);
   }

   /* cache the script name information */
   strcpy (captr->ScriptName, tkptr->ScriptName);
   strcpy (captr->ResFileName, tkptr->SearchOds.ResFileName);
   captr->ScriptRunTimePtr = tkptr->ScriptRunTimePtr;
   captr->HitCount = 0;
   captr->UsageCount++;
   memcpy (&captr->LastBinaryTime, &rqptr->rqTime.Vms64bit, 8);
}

/*****************************************************************************/
/*
Set all expanded file name entries in the script name cache to empty.  This
function is called when the DCLSupervisor() determines there are no more
zombies or CGIplus scripts active, and also when subprocesses are purged.
*/ 

DclPurgeScriptNameCache ()

{
   register struct DclScriptNameCacheStruct  *captr;
   register struct ListEntryStruct  *leptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DclPurgeScriptNameCache()\n");

   DclPurgeScriptNameCacheCount++;

   for (leptr = DclScriptNameCacheList.HeadPtr;
        leptr != NULL;
        leptr = leptr->NextPtr)
   {
      captr = (struct DclScriptNameCacheStruct*)leptr;
      captr->ResFileName[0] = '\0';
      captr->ScriptRunTimePtr = NULL;
   }
}

/*****************************************************************************/
/*
Perform an asynchronous search for the script file specified by
'tkptr->ScriptFileName'.  When found return the actual script file in
'tkptr->SearchOds.ResFileName' and call DclBeginScript().  If not found set to
empty and conclude the task.  The search is performed first by checking for a
file with the file type supplied with the script specification (usually not),
then by defaulting to ".COM", then ".EXE", and finally to any configured
runtime file types, in that order.

The file search uses a DCL task structure but no subprocess I/O is actually
underway. As the search is AST-driven it is possible for a cancelling client
to disconnect from the task structure via DclConcludeTask() while the search
is in progress. Each find-script-file function checks for the request pointer
and aborts the search if no longer present. The 'tkptr->FindScript' flag
indicates the structure is in use for this purpose and so will not be
reallocated until reset as the search is finally aborted.
*/ 

DclFindScript (struct DclTaskStruct *tkptr)

{
   register struct RequestStruct  *rqptr;

   int  status,
        idx;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "DclFindScript() %d |%s|\n",
               tkptr->FindScriptState,
               tkptr->ScriptFileName);

   if ((rqptr = tkptr->RequestPtr) == NULL)
   {
      /* request thread has disconnected during file search */
      DclFindScriptEnd (tkptr, true);
      return;
   }

   switch (tkptr->FindScriptState)
   {
      case DCL_FIND_SCRIPT_BEGIN :

         tkptr->FindScript = true;

         /* first default to DCL procedure (overridden if file type supplied) */
         OdsParse (&tkptr->SearchOds,
                   tkptr->ScriptFileName, 0, ".COM;", 5,
                   0, &DclFindScriptParseAst, tkptr);

         /* if not found then we'll be looking for executables */
         tkptr->FindScriptState = DCL_FIND_SCRIPT_EXE;

         return;

      case DCL_FIND_SCRIPT_EXE :

         /* move on to looking for executables */
         OdsParse (&tkptr->SearchOds,
                   tkptr->ScriptFileName, 0, ".EXE;", 5,
                   0, &DclFindScriptParseAst, tkptr);

         /* if not found then we'll be looking for runtime-defined */
         tkptr->FindScriptRunTimeIdx = 0;
         tkptr->FindScriptState = DCL_FIND_SCRIPT_RUNTIME;

         return;

      case DCL_FIND_SCRIPT_RUNTIME :

         /* looking through any list of user-defined script file types */
         idx = tkptr->FindScriptRunTimeIdx++;
         if (idx >= Config.cfScript.RunTimeCount)
         {
            /********************/
            /* script not found */
            /********************/

            /* indicate the script was not found */
            tkptr->SearchOds.ResFileName[0] = '\0';

            if (tkptr->CgiPlusCalloutFunction == NULL ||
                (void*)tkptr->CgiPlusCalloutFunction ==
                (void*)&DclCgiPlusDefaultCallout)
            {
               /* standard CGI/CGIplus script */
               rqptr->rqResponse.HttpStatus = 404;
               ErrorGeneral (rqptr, MsgFor(rqptr,MSG_SCRIPT_NOT_FOUND), FI_LI);
            }
            else
            {
               /* not a standard CGIplus script, an agent task */
               rqptr->rqResponse.HttpStatus = 500;
               ErrorGeneral (rqptr, "!AZ !AZ", tkptr->ScriptName, FI_LI);
            }

            DclFindScriptEnd (tkptr, true);

            return;
         }

         if (Debug)
            fprintf (stdout, "[%d]|%s|\n", idx, Config.cfScript.RunTimeArray[idx]);

         OdsParse (&tkptr->SearchOds,
                   tkptr->ScriptFileName, 0,
                   Config.cfScript.RunTimeArray[idx],
                   Config.cfScript.RunTimeLengthArray[idx],
                   0, &DclFindScriptParseAst, tkptr);

         return;

      default :
            ErrorExitVmsStatus (SS$_BUGCHECK, ErrorSanityCheck, FI_LI);
   }
}

/*****************************************************************************/
/*
AST called from DclFindScript() when asynchronous parse completes.  Check for
error status, then perform an asynchronous search with AST to
DclFindScriptSearchAst()
*/

DclFindScriptParseAst (struct FAB *FabPtr)

{
   register struct RequestStruct  *rqptr;
   register struct DclTaskStruct  *tkptr;

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout,
         "DclFindScriptParseAst() sts: %%X%08.08X stv: %%X%08.08X\n",
         FabPtr->fab$l_sts, FabPtr->fab$l_stv);

   tkptr = FabPtr->fab$l_ctx;

   if ((rqptr = tkptr->RequestPtr) == NULL)
   {
      /* request thread has disconnected during file search */
      DclFindScriptEnd (tkptr, true);
      return;
   }

   if (VMSnok (status = FabPtr->fab$l_sts))
   {
      rqptr->rqResponse.ErrorTextPtr = tkptr->ScriptName;
      rqptr->rqResponse.ErrorOtherTextPtr = tkptr->ScriptFileName;
      ErrorVmsStatus (rqptr, status, FI_LI);
      DclFindScriptEnd (tkptr, true);
      return;
   }

   if (Debug)
   {
      tkptr->SearchOds.NamVersionPtr[0] = '\0';
      fprintf (stdout, "|%s|\n", tkptr->SearchOds.ResFileName);
      tkptr->SearchOds.NamVersionPtr[0] = ';';
   }

   OdsSearch (&tkptr->SearchOds, &DclFindScriptSearchAst, tkptr);
}

/*****************************************************************************/
/*
AST called from DclFindScriptParseAst() when asynchronous search completes.
Check for a file-not-found status, if so then call DclFindScript() to search
for the next possibility.  Check for an error status.  If the file has been
found determine the runtime environment and initiate the subprocess execution.
*/

DclFindScriptSearchAst (struct FAB *FabPtr)

{
   register struct RequestStruct  *rqptr;
   register struct DclTaskStruct  *tkptr;

   register char  *cptr;
   int  status,
        idx;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout,
         "DclFindScriptSearchAst() sts: %%X%08.08X stv: %%X%08.08X\n",
         FabPtr->fab$l_sts, FabPtr->fab$l_stv);

   tkptr = FabPtr->fab$l_ctx;

   if ((rqptr = tkptr->RequestPtr) == NULL)
   {
      /* request thread has disconnected during file search */
      DclFindScriptEnd (tkptr, true);
      return;
   }

   if (VMSnok (status = FabPtr->fab$l_sts))
   {
      if (rqptr->WatchItem &&
          (WatchEnabled & WATCH_DCL))
      {
         tkptr->SearchOds.NamVersionPtr[0] = '\0';
         WatchThis (rqptr, FI_LI, WATCH_DCL,
                    "SEARCH !AZ %X!8XL %!%M",
                    tkptr->SearchOds.NamDevicePtr, status, status);
         tkptr->SearchOds.NamVersionPtr[0] = ';';
      }

      if (status == RMS$_FNF)
      {
         /* not found, look for another */
         DclFindScript (tkptr);
         return;
      }

      /* some other error, report it and finish up */
      rqptr->rqResponse.ErrorTextPtr = tkptr->ScriptName;
      rqptr->rqResponse.ErrorOtherTextPtr = tkptr->ScriptFileName;
      ErrorVmsStatus (rqptr, status, FI_LI);
      DclFindScriptEnd (tkptr, true);
      return;
   }

   if (rqptr->WatchItem &&
          (WatchEnabled & WATCH_DCL))
   {
      tkptr->SearchOds.NamVersionPtr[0] = '\0';
      WatchThis (rqptr, FI_LI, WATCH_DCL,
                 "SEARCH !AZ", tkptr->SearchOds.NamDevicePtr);
      tkptr->SearchOds.NamVersionPtr[0] = ';';
   }

   /*************/
   /* found it! */
   /*************/

   if (!memcmp (tkptr->SearchOds.NamTypePtr, ".COM;", 5))
      tkptr->ScriptRunTimePtr = "@";
   else
   if (!memcmp (tkptr->SearchOds.NamTypePtr, ".EXE;", 5))
      tkptr->ScriptRunTimePtr = "$";
   else
   {
      /* look through the list of user-definable script file types */
      for (idx = 0; idx < Config.cfScript.RunTimeCount; idx++)
      {
         if (!memcmp (tkptr->SearchOds.NamTypePtr,
                      Config.cfScript.RunTimeArray[idx],
                      Config.cfScript.RunTimeLengthArray[idx]))
            break;
      }
      if (idx < Config.cfScript.RunTimeCount)
      {
         /* found the file type (file extension) */
         for (cptr = Config.cfScript.RunTimeArray[idx];
              *cptr && *cptr != ' ';
              cptr++);
         if (*cptr) cptr++;
         tkptr->ScriptRunTimePtr = cptr;
      }
      else
      {
         /********************************************/
         /* don't know how to execute this file type */
         /********************************************/

         rqptr->rqResponse.HttpStatus = 500;
         ErrorGeneral (rqptr,
"Execution of &nbsp;<TT>!AZ</TT>&nbsp; script types not configured.",
                       tkptr->SearchOds.NamTypePtr, FI_LI);
         DclFindScriptEnd (tkptr, true);
         return;
      }
   }

   /**********************/
   /* execute the script */
   /**********************/

   /* terminate the expanded CGI file name */
   tkptr->SearchOds.NamVersionPtr[0] = '\0';
   strcpy (tkptr->ScriptFileName, tkptr->SearchOds.ResFileName);
   if (Debug) fprintf (stdout, "|%s|\n", tkptr->ScriptFileName);

   DclFindScriptEnd (tkptr, false);

   DclUpdateScriptNameCache (tkptr, rqptr);

   DclBeginScript (tkptr);
}

/*****************************************************************************/
/*
The parse structures only really need to be explicitly released if the script
is actually found, otherwise they are always implicitly released on the
sys$search() file-not-found! Here we'll do it all the time as we are dealing
with possible search abort as well as general search conclusion. When doing so
use scratch expanded file name space so as not to overwrite the CGI file name
if the script file was found.
*/

DclFindScriptEnd
(
struct DclTaskStruct *tkptr,
boolean FileNotFound
)
{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DclFindScriptEnd()\n");

   tkptr->FindScript = false;

   /* ensure parse internal data structures are released */
   if (tkptr->SearchOds.ParseInUse) OdsParseRelease (&tkptr->SearchOds);

   if (FileNotFound) DclConcludeTask (tkptr, false);
}

/*****************************************************************************/
/*
Called with an already-established CGIplus script, a DCL command, or when a
new CGIplus or CGI script file name has been searched for and found.  Begin
the I/O with the subprocess.
*/ 

DclBeginScript (struct DclTaskStruct *tkptr)

{
   register struct RequestStruct  *rqptr;

   int  status,
        Length;
   char  *ContentPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DclBeginScript()\n");

   if ((rqptr = tkptr->RequestPtr) == NULL)
   {
      /* request thread has disconnected during file search */
      DclConcludeTask (tkptr, false);
      return;
   }

   if (Debug)
      fprintf (stdout, "TaskType: %d DclCommand |%s|\n",
               tkptr->TaskType, tkptr->DclCommand);

   if (tkptr->TaskType == DCL_TASK_TYPE_CGI_SCRIPT ||
       tkptr->TaskType == DCL_TASK_TYPE_CGIPLUS_SCRIPT)
   {
      /* provide request stream (HTTP$INPUT) to subprocess */
      if (Debug) 
         fprintf (stdout, "|%d|%d|%s|\n",
                  rqptr->rqHeader.ContentLength,
                  rqptr->rqBody.CurrentCount,
                  rqptr->rqBody.BufferPtr);

      /* initialize content information (POST, PUT, etc., only) */
      tkptr->ContentPtr = rqptr->rqBody.BufferPtr;
      tkptr->ContentLength = rqptr->rqBody.CurrentCount;

      if (Config.cfScript.FullRequest)
      {
         /* supply request header and body (pre-4.3 behaviour) */

         if (rqptr->WatchItem &&
             (WatchEnabled & WATCH_DCL))
         {
            WatchThis (rqptr, FI_LI, WATCH_DCL,
                       "WRITE SYS$INPUT !UL bytes",
                       rqptr->rqHeader.RequestHeaderLength);
            WatchData (rqptr->rqHeader.RequestHeaderPtr,
                       rqptr->rqHeader.RequestHeaderLength);
         }

         if (VMSnok (status =
             sys$qio (0, tkptr->HttpInputChannel,
                      IO$_WRITELBLK, &tkptr->HttpInputIOsb,
                      &DclHttpInputAst, tkptr,
                      rqptr->rqHeader.RequestHeaderPtr, rqptr->rqHeader.RequestHeaderLength,
                      0, 0, 0, 0)))
         {
            /* report this error via the AST */
            tkptr->HttpInputIOsb.Status = status;
            SysDclAst (&DclHttpInputAst, tkptr);
         }
      }
      else
      if (tkptr->ContentLength)
      {
         /* supply only the body of the request (standard CGI behaviour) */
         if (tkptr->ContentLength > NetReadBufferSize)
            Length = NetReadBufferSize;
         else
            Length = tkptr->ContentLength;
         ContentPtr = tkptr->ContentPtr;
         tkptr->ContentPtr += Length;
         tkptr->ContentLength -= Length;

         if (rqptr->WatchItem &&
             (WatchEnabled & WATCH_DCL))
         {
            WatchThis (rqptr, FI_LI, WATCH_DCL,
                       "WRITE SYS$INPUT !UL bytes", Length);
            if (Length) WatchData (ContentPtr, Length);
         }

         if (VMSnok (status =
             sys$qio (0, tkptr->HttpInputChannel,
                      IO$_WRITELBLK, &tkptr->HttpInputIOsb,
                      &DclHttpInputAst, tkptr,
                      ContentPtr, Length,
                      0, 0, 0, 0)))
         {
            /* report this error via the AST */
            tkptr->HttpInputIOsb.Status = status;
            SysDclAst (&DclHttpInputAst, tkptr);
         }
      }
      else
      {
         /* the request has no body */

         if (rqptr->WatchItem &&
             (WatchEnabled & WATCH_DCL))
            WatchThis (rqptr, FI_LI, WATCH_DCL, "WRITE SYS$INPUT EOF");

         if (VMSnok (status =
             sys$qio (0, tkptr->HttpInputChannel,
                      IO$_WRITEOF, &tkptr->HttpInputIOsb,
                      &DclHttpInputAst, tkptr,
                      0, 0, 0, 0, 0, 0)))
         {
            /* report this error via the AST */
            tkptr->HttpInputIOsb.Status = status;
            SysDclAst (&DclHttpInputAst, tkptr);
         }
      }
      tkptr->QueuedHttpInput++;
   }

   switch (tkptr->TaskType)
   {
      case DCL_TASK_TYPE_CGI_SCRIPT :

         tkptr->QueuedSysCommandAllowed = 0;
         rqptr->rqResponse.ErrorTextPtr =
            MsgFor(rqptr,MSG_SCRIPT_DCL_ENVIRONMENT);
         if (VMSnok (DclCgiScriptSysCommand (tkptr)))
         {
            DclConcludeTask (tkptr, true);
            return;
         }

         break;

      case DCL_TASK_TYPE_CGIPLUS_SCRIPT :

         /* allow for the outstanding queued "STOP/ID=0" and EOF */
         tkptr->QueuedSysCommandAllowed = 2;
         if (tkptr->QueuedSysCommand < tkptr->QueuedSysCommandAllowed)
            DclCgiPlusScriptSysCommand (tkptr);

         /* here comes the CGI variable stream */
         if (VMSnok (DclCgiPlusScriptCgiPlusIn (tkptr)))
         {
            DclConcludeTask (tkptr, true);
            return;
         }

         break;

      case DCL_TASK_TYPE_CLI :

         tkptr->QueuedSysCommandAllowed = 0;
         if (VMSnok (DclCgiScriptSysCommand (tkptr)))
         {
            DclConcludeTask (tkptr, true);
            return;
         }

         break;

      default :
            ErrorExitVmsStatus (SS$_BUGCHECK, ErrorSanityCheck, FI_LI);
   }

   /* queue the initial read of the subprocess' SYS$OUTPUT */
   DclQioSysOutput (tkptr);

   /* remove default error message */
   rqptr->rqResponse.ErrorTextPtr = NULL;
}

/*****************************************************************************/
/*
Called if the HTTPd server is being restarted (image run down and then
restarted).  Does not need to be done if process is exiting.  This function
should be included in the image's onexit() function.
*/

DclExit ()

{
   register struct DclTaskStruct  *tkptr;
   register struct ListEntryStruct  *leptr;

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DclExit()\n");

   for (leptr = DclTaskList.HeadPtr; leptr != NULL; leptr = leptr->NextPtr)
   {
      tkptr = (struct DclTaskStruct*)leptr;
      if (!tkptr->SubprocessPid) continue;
      status = sys$delprc (&tkptr->SubprocessPid, 0);
      if (Debug)
         fprintf (stdout, "sys$delprc() %d %08.08X %%X%08.08X\n",
                  tkptr, tkptr->SubprocessPid, status);
   }
}

/*****************************************************************************/
/*
The maximum number of concurrent subprocesses has been reached. Look through
the DCL task structure for a CGIplus subprocess not currently in use. Find the
least used subprocess and delete it.  A lifetime count of -1 indicates the
subprocess has requested that it be immune to supervisor purging.
*/

DclSubprocessPurge ()

{
   register struct DclTaskStruct  *tkptr;
   register struct ListEntryStruct  *leptr;

   int  status,
        MinUsageCount;
   struct DclTaskStruct  *mintkptr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "DclSubprocessPurge() %d %d\n",
               DclSubprocessCount, DclSubprocessSoftLimit);

   DclPurgeCount++;
   MinUsageCount = 999999999;
   mintkptr = NULL;

   for (leptr = DclTaskList.HeadPtr; leptr != NULL; leptr = leptr->NextPtr)
   {
      tkptr = (struct DclTaskStruct*)leptr;

      if (Debug) DclTaskItemDebug (leptr, tkptr);

      if (!tkptr->SubprocessPid ||
          tkptr->TaskType != DCL_TASK_TYPE_CGIPLUS_SCRIPT ||
          tkptr->LifeTimeCount == -1 ||
          tkptr->QueuedSysOutput ||
          tkptr->QueuedCgiPlusIn ||
          tkptr->QueuedHttpInput ||
          tkptr->RequestPtr != NULL ||
          tkptr->FindScript ||
          tkptr->MarkedForDelete)
         continue;

      if (tkptr->CgiPlusUsageCount < MinUsageCount)
         MinUsageCount = (mintkptr = tkptr)->CgiPlusUsageCount;
   }

   if (mintkptr == NULL) return;

   mintkptr->MarkedForDelete = true;
   DclConcludeTask (mintkptr, true);
   DclSoftLimitPurgeCount++;
}

/*****************************************************************************/
/*
This function may be called at any stage to rundown or abort a DCL task,
including during the search for the script file (see note in DclFindScript()).
If there is still outstanding I/O this is cancelled as appropriate to task
rundown or abort. If no outstanding I/O then if there is an associated request
that request's next task function is called.
*/

DclConcludeTask
(
struct DclTaskStruct *tkptr,
boolean AbortTask
)
{
   int  status;
   char  ScriptFileName [ODS_MAX_FILE_NAME_LENGTH+1],
         ScriptName [SCRIPT_NAME_SIZE];
   void  *CgiPlusCalloutFunction,
         *NextTaskFunction;
   struct RequestStruct  *RequestPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
   {
      fprintf (stdout, "DclConcludeTask() %d %d %d %d %d %d %d %08.08X\n",
         AbortTask, tkptr, tkptr->RequestPtr, tkptr->QueuedSysCommand,
         tkptr->QueuedSysOutput, tkptr->QueuedCgiPlusIn,
         tkptr->QueuedHttpInput, tkptr->SubprocessPid);
   }

   if (AbortTask)
   {
      tkptr->MarkedForDelete = true;
      tkptr->QueuedSysCommandAllowed = 0;
   }

   if (tkptr->QueuedSysCommand > tkptr->QueuedSysCommandAllowed ||
        tkptr->QueuedSysOutput ||
        tkptr->QueuedCgiPlusIn ||
        tkptr->QueuedHttpInput)
   {
      /****************/
      /* rundown task */
      /****************/

      if (AbortTask)
      {
         /* cancel everything everywhere! */
         sys$cancel (tkptr->SysCommandChannel);
         sys$cancel (tkptr->SysOutputChannel);
         sys$cancel (tkptr->CgiPlusInChannel);
         sys$cancel (tkptr->HttpInputChannel);
      }
      else
      {
         /* cancel any outstanding I/O on input channels */
         if (tkptr->QueuedCgiPlusIn) sys$cancel (tkptr->CgiPlusInChannel);
         if (tkptr->QueuedHttpInput) sys$cancel (tkptr->HttpInputChannel);
         if (tkptr->QueuedSysCommand > tkptr->QueuedSysCommandAllowed)
            sys$cancel (tkptr->SysCommandChannel);
      }

      return;
   }

   /*********************/
   /* task has finished */
   /*********************/

   if (tkptr->ScriptName[0] && !tkptr->SearchOds.ResFileName[0])
   {
      /* search for script file was unsuccessful */
      if (tkptr->RequestPtr != NULL)
      {
         /* still has an associated request, declare the next task */
         SysDclAst (tkptr->NextTaskFunction, tkptr->RequestPtr);

         /* disassociate the DCL task and request structures */
         tkptr->RequestPtr->DclTaskPtr = NULL;
         tkptr->RequestPtr = tkptr->NextTaskFunction = NULL;
         tkptr->WatchItem = 0;
      }

      if (tkptr->TaskType == DCL_TASK_TYPE_CGIPLUS_SCRIPT)
         tkptr->MarkedForDelete = true;
   }
   else
   if (tkptr->RequestPtr != NULL)
   {
      if (!tkptr->SubprocessActivated &&
          tkptr->RequestPtr->rqCgi.ScriptRetryCount++ < DclSubprocessSoftLimit)
      {
         /* looks like the script exited before actually starting! */
         tkptr->MarkedForDelete = true;

         /* retrieve necessary information from the current DCL task */
         RequestPtr = tkptr->RequestPtr;
         NextTaskFunction = tkptr->NextTaskFunction;
         strcpy (ScriptName, tkptr->ScriptName);
         strcpy (ScriptFileName, tkptr->ScriptFileName);
         CgiPlusCalloutFunction = tkptr->CgiPlusCalloutFunction;

         /* disassociate the DCL task and request structures */
         tkptr->RequestPtr->DclTaskPtr = NULL;
         tkptr->RequestPtr = tkptr->NextTaskFunction = NULL;
         tkptr->WatchItem = 0;

         /* restart using a new DCL task */
         DclBegin (RequestPtr, NextTaskFunction, NULL, ScriptName, NULL,
                   ScriptFileName, NULL);
      }
      else
      {
         if (!tkptr->SubprocessResponded &&
             (tkptr->TaskType == DCL_TASK_TYPE_CGI_SCRIPT ||
              tkptr->TaskType == DCL_TASK_TYPE_CGIPLUS_SCRIPT))
         {
            /* hmmm, script has not provided any output! */
            if (tkptr->RequestPtr->rqHeader.Method == HTTP_METHOD_GET)
            {
               /* blame script for general GET method failures */
               tkptr->RequestPtr->rqResponse.HttpStatus = 502;
               ErrorGeneral (tkptr->RequestPtr,
                  MsgFor(tkptr->RequestPtr,MSG_SCRIPT_RESPONSE_ERROR), FI_LI);
            }
            else
            {
               /* other methods are probably not implemented by the script */
               tkptr->RequestPtr->rqResponse.HttpStatus = 501;
               ErrorGeneral (tkptr->RequestPtr,
                  MsgFor(tkptr->RequestPtr,MSG_REQUEST_METHOD), FI_LI);
            }
         }

         /* still has an associated request, declare the next task */
         SysDclAst (tkptr->NextTaskFunction, tkptr->RequestPtr);

         /* disassociate the DCL task and request structures */
         tkptr->RequestPtr->DclTaskPtr = NULL;
         tkptr->RequestPtr = tkptr->NextTaskFunction = NULL;
         tkptr->WatchItem = 0;
      }
   }

   if (DclUseZombies ||
       tkptr->TaskType == DCL_TASK_TYPE_CGIPLUS_SCRIPT)
   {
      /* empty any remains in sys$output! */
      DclEmptySysOutput (tkptr);
   }

   /* if marked for process deletion */
   if (tkptr->MarkedForDelete)
   {
      tkptr->MarkedForDelete = false;
      if (tkptr->SubprocessPid)
      {
         status = sys$delprc (&tkptr->SubprocessPid, 0);
         if (Debug)
            fprintf (stdout, "sys$delprc() %08.08X %%X%08.08X\n",
                     tkptr->SubprocessPid, status);
      }
   }

   if (DclUseZombies &&
       tkptr->TaskType != DCL_TASK_TYPE_CGIPLUS_SCRIPT)
      tkptr->ZombieCount++;
}

/*****************************************************************************/
/*
If an error is encountered an error message is generated and the error status
returned.  It is up to the calling routine to abort the processing.  Queue a
writer-wait I/O to the SYS$OUTPUT channel to stall I/O until the subprocess
has started.
*/ 

DclCreateSubprocess (struct DclTaskStruct *tkptr)

{
   /* spawn DCL subprocess with account's authorized privileges */
   static unsigned long  /* authpriv,trusted,nowait,noclisym,nolognam,nokey */
                         AuthPrivSpawnFlags = 0xcf;

   /* spawn DCL subprocess with current privileges (none at all!) */
   static unsigned long  /* nowait,noclisym,nolognam,nokey */
                         CurrPrivSpawnFlags = 0x0f;

   static int  SubprocessNumber;
   static char  SubprocessName [16];
   static $DESCRIPTOR (SubprocessNameFaoDsc, "HTTPd:!UL-!UL");
   static $DESCRIPTOR (SubprocessNameDsc, SubprocessName);

   register struct RequestStruct  *rqptr;

   int  status,
        Count,
        DclAstStatus;
   unsigned short  Length;
   unsigned long  SpawnFlags;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DclCreateSubprocess()\n");

   /* get a pointer to the request structure */
   rqptr = tkptr->RequestPtr;

   if (Config.cfScript.SpawnAuthPriv)
      SpawnFlags = AuthPrivSpawnFlags;
   else
      SpawnFlags = CurrPrivSpawnFlags;

#if OPERATE_WITH_SYSPRV
   /* if we're normally operating with SYSPRV then turn off before spawning */
   if (OperateWithSysPrv)
      if (VMSnok (status = sys$setprv (0, &SysPrvMask, 0, 0)))
         ErrorExitVmsStatus (status, "sys$setprv()", FI_LI);
#endif

   for (Count = 100; Count; Count--)
   {
      /* subprocesses are consecutively numbered from 1..999 */
      SubprocessNameDsc.dsc$w_length = sizeof(SubprocessName)-1;
      sys$fao (&SubprocessNameFaoDsc, &Length, &SubprocessNameDsc, 
               ServerPort, SubprocessNumber++ % 1000 + 1);
      SubprocessName[SubprocessNameDsc.dsc$w_length = Length] = '\0';
      if (Debug) fprintf (stdout, "SubprocessName |%s|\n", SubprocessName);

      status = lib$spawn (0,
                          &tkptr->SysCommandDevNameDsc,
                          &tkptr->SysOutputDevNameDsc,
                          &SpawnFlags,
                          &SubprocessNameDsc,
                          &tkptr->SubprocessPid,
                          &tkptr->SubprocessCompletionStatus,
                          0,
                          &DclSubprocessCompletionAST, tkptr,
                          0, 0, 0);
      if (Debug)
         fprintf (stdout, "lib$spawn() %%X%08.08X PID: %08.08X\n",
                  status, tkptr->SubprocessPid);

      if (status != SS$_DUPLNAM) break;
   }

#if OPERATE_WITH_SYSPRV
   /* if we're normally operating with SYSPRV then turn back on after */
   if (OperateWithSysPrv)
      if (VMSnok (status = sys$setprv (1, &SysPrvMask, 0, 0)))
         ErrorExitVmsStatus (status, "sys$setprv()", FI_LI);
#endif

   if (VMSok (status))
   {
      Accounting.DclSpawnCount++;
      return (status);
   }

   rqptr->rqResponse.ErrorTextPtr = MsgFor(rqptr,MSG_SCRIPT_SPAWN);
   ErrorVmsStatus (rqptr, status, FI_LI);
   return (status);
}

/*****************************************************************************/
/*
This AST is called when the subprocesses exits.
*/

DclSubprocessCompletionAST (struct DclTaskStruct *tkptr)

{
   /*********/
   /* begin */
   /*********/

   if (Debug)
   {
      fprintf (stdout,
         "DclSubprocessCompletionAST() %08.08X %%X%08.08X\n",
         tkptr->SubprocessPid, tkptr->SubprocessCompletionStatus);
   }

   if (WatchEnabled & WATCH_DCL)
   {
      if (VMSok (tkptr->SubprocessCompletionStatus))
         WatchThis (tkptr->RequestPtr, FI_LI, WATCH_DCL,
                    "SUBPROCESS completion pid !8XL", tkptr->SubprocessPid);
      else
         WatchThis (tkptr->RequestPtr, FI_LI, WATCH_DCL,
                    "SUBPROCESS completion pid !8XL %X!8XL %!%M",
                    tkptr->SubprocessPid,
                    tkptr->SubprocessCompletionStatus,
                    tkptr->SubprocessCompletionStatus);
   }

   if (tkptr->TaskType == DCL_TASK_TYPE_CGIPLUS_SCRIPT)
   {
      /* ensure SYS$COMMAND gets emptied! */
      tkptr->QueuedSysCommandAllowed = 0;
   }

   /* set the PID and lifetime to zero */
   tkptr->SubprocessPid = tkptr->LifeTimeCount = 0;

   /* keep track of how many subprocesses are executing */
   if (DclSubprocessCount) DclSubprocessCount--;

   /* ensure any old sequence strings are not reused */
   tkptr->CgiEof[0] =
      tkptr->CgiEot[0] =
      tkptr->CgiEsc[0] = '\0';
   tkptr->CgiEofLength =
      tkptr->CgiEotLength =
      tkptr->CgiEscLength = 0;

   DclConcludeTask (tkptr, false);
}

/*****************************************************************************/
/*
If a script application writes any output before it was/is terminated that
output will still be laying unread in the mailbox.  Clear any such noise lest
it be read by the next subprocess to use the mailbox.  Does this synchronously
(i.e. with QIO waits).
*/

DclEmptySysOutput (struct DclTaskStruct *tkptr)

{
   struct MbxSenseIOsb  SenseIOsb;
   int  LastMessageCount;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DclEmptySysOutput()\n");

   LastMessageCount = 999999999;
   SenseIOsb.MessageCount = 0;

   for (;;)
   {
      sys$qiow (0, tkptr->SysOutputChannel,
                IO$_SENSEMODE, &SenseIOsb,
                0, 0, 0, 0, 0, 0, 0, 0);

      if (VMSnok (SenseIOsb.Status))
         ErrorExitVmsStatus (SenseIOsb.Status, "sys$qiow()", FI_LI);

      if (Debug)
         fprintf (stdout, "sys$qiow() IO$_SENSEMODE %%X%08.08X %d %d\n", 
                  SenseIOsb.Status, SenseIOsb.MessageCount,
                  SenseIOsb.MessageBytes);

      if (!SenseIOsb.MessageCount) break;
      /* potential infinite loop, check message count is decreasing! */
      if (LastMessageCount <= SenseIOsb.MessageCount) break;
      LastMessageCount = SenseIOsb.MessageCount;

      sys$qiow (0, tkptr->SysOutputChannel,
                IO$_READLBLK | IO$M_WRITERCHECK,
                &tkptr->SysOutputIOsb, 0, 0,
                tkptr->SysOutputPtr, tkptr->SysOutputSize, 0, 0, 0, 0);

      if (Debug)
      {
         tkptr->SysOutputPtr[tkptr->SysOutputIOsb.Count] = '\0';
         fprintf (stdout, "sys$qiow() IO$_READLBLK %%X%08.08X %d\n|%s|\n", 
                  tkptr->SysOutputIOsb.Status, tkptr->SysOutputIOsb.Count,
                  tkptr->SysOutputPtr);
      }
   }
}

/*****************************************************************************/
/*
If an error is encountered an error message is generated and the error status
returned.  It is up to the calling routine to abort the processing.  Create
four mailboxes that will be associated with the subprocess I/O streams.  If
an error occurs any mailbox created up to that point is deleted and the
channel set back to zero.
*/ 

DclCreateMailboxes (struct DclTaskStruct *tkptr)

{
   static unsigned long  DevNamItem = DVI$_DEVNAM;

   int  status;
   unsigned short  Length;
   unsigned long  BytLmAfter,
                  BytLmBefore;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DclCreateMailboxes()\n");

   BytLmBefore = GetJpiBytLm ();

   if (Debug)
      fprintf (stdout, "DclMailboxBytLmRequired: %d BytLmBefore: %d\n",
              DclMailboxBytLmRequired, BytLmBefore);

   /* ensure we're leaving enough BYTLM for client socket creation at least */
   if (DclMailboxBytLmRequired &&
       BytLmBefore - DclMailboxBytLmRequired <=
       NetAcceptBytLmRequired * Config.cfServer.BusyLimit)
   {
      ErrorNoticed (0, "BYTLM exhausted", FI_LI);
      return (SS$_EXQUOTA);
   }

   /***********************/
   /* SYS$COMMAND mailbox */
   /***********************/

   if (VMSnok (status =
       sys$crembx (0,
                   &tkptr->SysCommandChannel,
                   DclSysCommandSize, DclSysCommandSize,
                   SubprocessMbxProtectionMask,
                   0, 0, CMB$M_WRITEONLY)))
   {
      if (Debug) fprintf (stdout, "sys$crembx() %%X%08.08X\n", status);
      return (status);
   }

   tkptr->SysCommandDevNameDsc.dsc$w_length = sizeof(tkptr->SysCommandDevName);
   tkptr->SysCommandDevNameDsc.dsc$a_pointer = tkptr->SysCommandDevName;
   tkptr->SysCommandDevNameDsc.dsc$b_class = DSC$K_CLASS_S;
   tkptr->SysCommandDevNameDsc.dsc$b_dtype = DSC$K_DTYPE_T;

   if (VMSnok (status =
       lib$getdvi (&DevNamItem, &tkptr->SysCommandChannel,
                   0, 0, &tkptr->SysCommandDevNameDsc, &Length)))
   {
      sys$dassgn (tkptr->SysCommandChannel);
      tkptr->SysCommandChannel = 0;
      return (status);
   }
   tkptr->SysCommandDevName
      [tkptr->SysCommandDevNameDsc.dsc$w_length = Length] = '\0';
   if (Debug)
      fprintf (stdout, "SysCommandDevName |%s|\n", tkptr->SysCommandDevName);

   /**********************/
   /* SYS$OUTPUT mailbox */
   /**********************/

   if (VMSnok (status =
       sys$crembx (0,
                   &tkptr->SysOutputChannel,
                   DclSysOutputSize, DclSysOutputSize,
                   SubprocessMbxProtectionMask,
                   0, 0, CMB$M_READONLY)))
   {
      if (Debug) fprintf (stdout, "sys$crembx() %%X%08.08X\n", status);
      sys$dassgn (tkptr->SysCommandChannel);
      tkptr->SysCommandChannel = 0;
      return (status);
   }

   tkptr->SysOutputDevNameDsc.dsc$w_length = sizeof(tkptr->SysOutputDevName);
   tkptr->SysOutputDevNameDsc.dsc$a_pointer = tkptr->SysOutputDevName;
   tkptr->SysOutputDevNameDsc.dsc$b_class = DSC$K_CLASS_S;
   tkptr->SysOutputDevNameDsc.dsc$b_dtype = DSC$K_DTYPE_T;

   if (VMSnok (status =
       lib$getdvi (&DevNamItem, &tkptr->SysOutputChannel,
                   0, 0, &tkptr->SysOutputDevNameDsc, &Length)))
   {
      sys$dassgn (tkptr->SysCommandChannel);
      tkptr->SysCommandChannel = 0;
      sys$dassgn (tkptr->SysOutputChannel);
      tkptr->SysOutputChannel = 0;
      tkptr->HttpInputChannel = tkptr->SysCommandChannel = 0;
      return (status);
   }
   tkptr->SysOutputDevName
      [tkptr->SysOutputDevNameDsc.dsc$w_length = Length] = '\0';
   if (Debug)
      fprintf (stdout, "SysOutputDevName |%s|\n", tkptr->SysOutputDevName);

   /*********************/
   /* CGIPLUSIN mailbox */
   /*********************/

   if (VMSnok (status =
       sys$crembx (0,
                   &tkptr->CgiPlusInChannel,
                   DclCgiPlusInSize, DclCgiPlusInSize,
                   SubprocessMbxProtectionMask,
                   0, 0, CMB$M_WRITEONLY)))
   {
      sys$dassgn (tkptr->SysCommandChannel);
      tkptr->SysCommandChannel = 0;
      sys$dassgn (tkptr->SysOutputChannel);
      tkptr->SysOutputChannel = 0;
      if (Debug) fprintf (stdout, "sys$crembx() %%X%08.08X\n", status);
      return (status);
   }

   tkptr->CgiPlusInDevNameDsc.dsc$w_length = sizeof(tkptr->CgiPlusInDevName);
   tkptr->CgiPlusInDevNameDsc.dsc$a_pointer = tkptr->CgiPlusInDevName;
   tkptr->CgiPlusInDevNameDsc.dsc$b_class = DSC$K_CLASS_S;
   tkptr->CgiPlusInDevNameDsc.dsc$b_dtype = DSC$K_DTYPE_T;

   if (VMSnok (status =
       lib$getdvi (&DevNamItem, &tkptr->CgiPlusInChannel,
                   0, 0, &tkptr->CgiPlusInDevNameDsc, &Length)))
   {
      sys$dassgn (tkptr->SysCommandChannel);
      tkptr->SysCommandChannel = 0;
      sys$dassgn (tkptr->SysOutputChannel);
      tkptr->SysOutputChannel = 0;
      sys$dassgn (tkptr->CgiPlusInChannel);
      tkptr->CgiPlusInChannel = 0;
      return (status);
   }
   tkptr->CgiPlusInDevName
     [tkptr->CgiPlusInDevNameDsc.dsc$w_length = Length] = '\0';
   if (Debug)
      fprintf (stdout, "CgiPlusInDevName |%s|\n", tkptr->CgiPlusInDevName);

   /**********************/
   /* HTTP$INPUT mailbox */
   /**********************/

   if (VMSnok (status =
       sys$crembx (0,
                   &tkptr->HttpInputChannel,
                   NetReadBufferSize, NetReadBufferSize,
                   SubprocessMbxProtectionMask,
                   0, 0, CMB$M_WRITEONLY)))
   {
      if (Debug) fprintf (stdout, "sys$crembx() %%X%08.08X\n", status);
      sys$dassgn (tkptr->SysCommandChannel);
      tkptr->SysCommandChannel = 0;
      sys$dassgn (tkptr->SysOutputChannel);
      tkptr->SysOutputChannel = 0;
      sys$dassgn (tkptr->CgiPlusInChannel);
      tkptr->CgiPlusInChannel = 0;
      return (status);
   }

   tkptr->HttpInputDevNameDsc.dsc$w_length = sizeof(tkptr->HttpInputDevName);
   tkptr->HttpInputDevNameDsc.dsc$a_pointer = tkptr->HttpInputDevName;
   tkptr->HttpInputDevNameDsc.dsc$b_class = DSC$K_CLASS_S;
   tkptr->HttpInputDevNameDsc.dsc$b_dtype = DSC$K_DTYPE_T;

   if (VMSnok (status =
       lib$getdvi (&DevNamItem, &tkptr->HttpInputChannel,
                   0, 0, &tkptr->HttpInputDevNameDsc, &Length)))
   {
      sys$dassgn (tkptr->SysCommandChannel);
      tkptr->SysCommandChannel = 0;
      sys$dassgn (tkptr->SysOutputChannel);
      tkptr->SysOutputChannel = 0;
      sys$dassgn (tkptr->CgiPlusInChannel);
      tkptr->CgiPlusInChannel = 0;
      sys$dassgn (tkptr->HttpInputChannel);
      tkptr->HttpInputChannel = 0;
      return (status);
   }
   tkptr->HttpInputDevName
     [tkptr->HttpInputDevNameDsc.dsc$w_length = Length] = '\0';
   if (Debug)
      fprintf (stdout, "HttpInputDevName |%s|\n", tkptr->HttpInputDevName);

   if (!DclMailboxBytLmRequired)
   {
      BytLmAfter = GetJpiBytLm ();
      DclMailboxBytLmRequired = BytLmBefore - BytLmAfter;
      if (Debug)
         fprintf (stdout, "DclMailboxBytLmRequired: %d\n",
                  DclMailboxBytLmRequired);
   }

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
Queue up a read from the subprocess "SYS$OUTPUT" mailbox.  When the read 
completes call function DclSysOutputAst(), do any post-processing 
required and write the data to the client over the network.  The next read 
from the subprocess via the mailbox will be queued by the network write 
completion AST function.
*/ 

DclQioSysOutput (struct DclTaskStruct *tkptr)

{
   int  status,
        DclAstStatus;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "DclSysOutput() %d %d\n", tkptr, tkptr->RequestPtr);

   if (VMSnok (status =
       sys$qio (0, tkptr->SysOutputChannel,
                IO$_READLBLK | IO$M_WRITERCHECK,
                &tkptr->SysOutputIOsb,
                &DclSysOutputAst, tkptr,
                tkptr->SysOutputPtr, tkptr->SysOutputSize, 0, 0, 0, 0)))
   {
      /* report this error via the AST */
      tkptr->SysOutputIOsb.Status = status;
      SysDclAst (&DclSysOutputAst, tkptr);
   }
   tkptr->QueuedSysOutput++;
}

/*****************************************************************************/
/*
A queued asynchronous read from the subprocess "SYS$OUTPUT" mailbox has 
completed. 
*/ 

DclSysOutputAst (struct DclTaskStruct *tkptr)

{
   int  cnt, status, value;
   char  *cptr;
   struct RequestStruct  *rqptr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
   {
      fprintf (stdout,
         "DclSysOutputAst() %d %d %08.08X %d %d %d %d %d %%X%08.08X\n",
         tkptr, tkptr->RequestPtr, tkptr->SubprocessPid,
         tkptr->QueuedSysCommand, tkptr->QueuedSysOutput,
         tkptr->QueuedCgiPlusIn, tkptr->QueuedHttpInput,
         tkptr->SysOutputIOsb.Count, tkptr->SysOutputIOsb.Status);

/*
      tkptr->SysOutputPtr[tkptr->SysOutputIOsb.Count] = '\0';
      fprintf (stdout, "|%s|\n", tkptr->SysOutputPtr);
*/
   }

   if (tkptr->QueuedSysOutput) tkptr->QueuedSysOutput--;

   rqptr = tkptr->RequestPtr;

   if (tkptr->WatchItem &&
       (WatchEnabled & WATCH_DCL))
   {
      if (VMSnok (tkptr->SysOutputIOsb.Status))
         WatchThis (rqptr, FI_LI, WATCH_DCL,
                    "READ SYS$OUTPUT %X!8XL %!%M",
                    tkptr->SysOutputIOsb.Status,
                    tkptr->SysOutputIOsb.Status);
      else
      {
         WatchThis (rqptr, FI_LI, WATCH_DCL,
                    "READ SYS$OUTPUT %X!8XL !UL bytes",
                    tkptr->SysOutputIOsb.Status,
                    tkptr->SysOutputIOsb.Count);
         if (tkptr->SysOutputIOsb.Count)
            WatchDataDump (tkptr->SysOutputPtr,
                           tkptr->SysOutputIOsb.Count);
      }
   }

   if (tkptr->SysOutputIOsb.Status == SS$_ENDOFFILE)
   {
      /* 
          If a script spawns multiple subprocesses each will
          terminate by queueing an end-of-file.  Ignore these.
          Queue the next read of the subprocess' SYS$OUTPUT, a
          SS$_NOWRITER or DCL EOF signals completion of output!
      */
      DclQioSysOutput (tkptr);
      return;
   }

   if (VMSnok (tkptr->SysOutputIOsb.Status))
   {
      if (tkptr->SysOutputIOsb.Status == SS$_NOWRITER)
      {
         /* subprocess has terminated */
         DclConcludeTask (tkptr, false);
         return;
      }

      if (tkptr->SysOutputIOsb.Status == SS$_ABORT ||
          tkptr->SysOutputIOsb.Status == SS$_CANCEL)
      {
         DclConcludeTask (tkptr, true);
         return;
      }

      rqptr->rqResponse.ErrorTextPtr = MsgFor(rqptr,MSG_SCRIPT_IPC);
      ErrorVmsStatus (rqptr, tkptr->SysOutputIOsb.Status, FI_LI);
      DclConcludeTask (tkptr, true);
      return;
   }

   /*********************/
   /* finally it's true */
   /*********************/

   tkptr->SubprocessResponded = true;

   tkptr->SysOutputPtr[tkptr->SysOutputIOsb.Count] = '\0';

   value = CgiOutput (rqptr, tkptr->SysOutputPtr, tkptr->SysOutputIOsb.Count);

   switch (value)
   {
      case CGI_OUTPUT_END :

         /* terminate processing */
         DclConcludeTask (tkptr, false);

         if (rqptr->rqCgi.CalloutInProgress)
         {
            if (tkptr->WatchItem &&
                (WatchEnabled & WATCH_DCL))
               WatchThis (rqptr, FI_LI, WATCH_DCL, "CALLOUT end");

            /* script didn't send an EOT, let callable function know now! */
            rqptr->rqCgi.CalloutInProgress = false;

            /* the NULL and zero bytes indicates transaction end */
            rqptr->rqCgi.CalloutOutputPtr = NULL;
            rqptr->rqCgi.CalloutOutputCount = 0;

            /* execute the callout function */
            (*tkptr->CgiPlusCalloutFunction)(rqptr);
         }

         if ((void*)tkptr->CgiPlusCalloutFunction !=
             (void*)&DclCgiPlusDefaultCallout)
         {
            /* agent task */
            if (rqptr->rqCgi.BufferLength)
            {
               /* release the CGI variables used for this task */
               if (rqptr->rqCgi.BufferPtr != NULL)
                  VmFreeFromHeap (rqptr, rqptr->rqCgi.BufferPtr, FI_LI); 
               rqptr->rqCgi.BufferLength = 0;
               rqptr->rqCgi.BufferPtr = rqptr->rqCgi.BufferCurrentPtr = NULL;
            }
         }

         return;

      case CGI_OUTPUT_NOT_STRICT :

         /* not strictly CGI compliant, report and terminate processing */
         tkptr->RequestPtr->rqResponse.HttpStatus = 502;
         ErrorGeneral (tkptr->RequestPtr,
            MsgFor(tkptr->RequestPtr,MSG_SCRIPT_RESPONSE_ERROR), FI_LI);
         DclConcludeTask (tkptr, true);
         return;

      case CGI_OUTPUT_ESCAPE_BEGIN :

         /* start talking direct to the server */
         if (tkptr->WatchItem &&
             (WatchEnabled & WATCH_DCL))
            WatchThis (rqptr, FI_LI, WATCH_DCL, "CALLOUT begin");

         rqptr->rqCgi.CalloutInProgress = true;

         /* the NULL and minus one bytes indicates transaction begin */
         rqptr->rqCgi.CalloutOutputPtr = NULL;
         rqptr->rqCgi.CalloutOutputCount = -1;

         /* execute the callout function */
         (*tkptr->CgiPlusCalloutFunction)(rqptr);

         DclQioSysOutput (tkptr);
         return;

      case CGI_OUTPUT_ESCAPE :

         /* the script is talking direct to the server */
         rqptr->rqCgi.CalloutOutputPtr = tkptr->SysOutputPtr;
         rqptr->rqCgi.CalloutOutputCount = tkptr->SysOutputIOsb.Count;

         if (tkptr->WatchItem &&
             (WatchEnabled & WATCH_DCL))
         {
            WatchThis (rqptr, FI_LI, WATCH_DCL,
                       "CALLOUT !UL bytes",
                       rqptr->rqCgi.CalloutOutputCount);
            if (rqptr->rqCgi.CalloutOutputCount)
               WatchDataDump (rqptr->rqCgi.CalloutOutputPtr,
                              rqptr->rqCgi.CalloutOutputCount);
         }

         /* execute the callout function */
         (*tkptr->CgiPlusCalloutFunction)(rqptr);

         DclQioSysOutput (tkptr);
         return;

      case CGI_OUTPUT_ESCAPE_END :

         if (tkptr->WatchItem &&
             (WatchEnabled & WATCH_DCL))
            WatchThis (rqptr, FI_LI, WATCH_DCL, "CALLOUT end");

         /* end talking direct to the server */
         rqptr->rqCgi.CalloutInProgress = false;

         /* the NULL and zero bytes indicates transaction end */
         rqptr->rqCgi.CalloutOutputPtr = NULL;
         rqptr->rqCgi.CalloutOutputCount = 0;

         /* execute the callout function */
         (*tkptr->CgiPlusCalloutFunction)(rqptr);

         DclQioSysOutput (tkptr);
         return;

      default :

         cptr = tkptr->SysOutputPtr;
         cnt = tkptr->SysOutputIOsb.Count;

         if (value > 0)
         {
            if (Debug) fprintf (stdout, "value: %d\n", value);
            cptr += value;
            cnt -= value;
         }

         /* step along any carriage control that may have been appended */
         while (cptr[cnt]) cnt++;

         if (!cnt)
         {
            /* output all absorbed, just queue the next read */
            DclQioSysOutput (tkptr);
            return;
         }

         if (rqptr->rqCgi.BufferRecords)
         {
            /* buffer the record */
            NetWriteBuffered (rqptr, &DclSysOutputToClientAst, cptr, cnt);
            tkptr->QueuedSysOutput++;
            return;
         }

         /* write the record */
         NetWrite (rqptr, &DclSysOutputToClientAst, cptr, cnt);
         tkptr->QueuedSysOutput++;
         return;
   }
}

/*****************************************************************************/
/*
************
*** NOTE ***  This function takes a pointer to a request!!!
************  Due to is being an AST from a general network write function.

A queued asynchronous write of subprocess SYS$OUTPUT (mailbox) to the client
over the network has completed.
*/ 

DclSysOutputToClientAst (struct RequestStruct *rqptr)

{
   register struct DclTaskStruct  *tkptr;

   int  status;

   /*********/
   /* begin */
   /*********/

   /* get a pointer to the DCL task from the request structure */
   tkptr = rqptr->DclTaskPtr;

   if (Debug)
   {
      fprintf (stdout,
         "DclSysOutputToClientAst() %d %d %d %d %d %d %%X%08.08X %%X%08.08X\n",
         tkptr, tkptr->RequestPtr, tkptr->QueuedSysCommand,
         tkptr->QueuedSysOutput, tkptr->QueuedCgiPlusIn,
         tkptr->QueuedHttpInput, tkptr->SysOutputIOsb.Status,
         rqptr->rqNet.WriteIOsb.Status);
   }

   if (tkptr->QueuedSysOutput) tkptr->QueuedSysOutput--;

   /* abort task if NETWORK ERROR when writing TO CLIENT */
   if (VMSnok (rqptr->rqNet.WriteIOsb.Status))
   {
      DclConcludeTask (tkptr, true);
      return;
   }

   /* queue the next read of the subprocess' SYS$OUTPUT */
   DclQioSysOutput (tkptr);
}

/*****************************************************************************/
/*
Queue up a write of data to the subprocess "SYS$COMMAND" mailbox.  This is the 
subprocesses' "SYS$COMMAND", supplying the DCL commands to execute, CGI
information, etc.
*/ 

DclQioSysCommand 
(
struct DclTaskStruct *tkptr,
char *DataPtr,
int DataLength
)
{
   int  status,
        DclAstStatus;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "DclQioSysCommand() %d %d |%s|\n",
               tkptr, tkptr->RequestPtr, DataPtr);

   if (DataPtr == NULL)
   {
      /* NULL pointer means write an end-of-file to the channel */

      if (tkptr->WatchItem &&
          (WatchEnabled & WATCH_DCL))
         WatchThis (tkptr->RequestPtr, FI_LI, WATCH_DCL,
                    "WRITE SYS$COMMAND EOF");

      status = sys$qio (0, tkptr->SysCommandChannel,
                        IO$_WRITEOF, &tkptr->SysCommandIOsb,
                        &DclSysCommandAst, tkptr,
                        0, 0, 0, 0, 0, 0);
   }
   else
   {
      if (tkptr->WatchItem &&
          (WatchEnabled & WATCH_DCL))
      {
         WatchThis (tkptr->RequestPtr, FI_LI, WATCH_DCL,
                    "WRITE SYS$COMMAND !UL bytes", DataLength);
         if (DataLength) WatchData (DataPtr, DataLength);
      }

      status = sys$qio (0, tkptr->SysCommandChannel,
                        IO$_WRITELBLK, &tkptr->SysCommandIOsb,
                        &DclSysCommandAst, tkptr,
                        DataPtr, DataLength, 0, 0, 0, 0);
   }

   if (VMSnok (status))
   {
      /* report this error via the AST */
      tkptr->SysCommandIOsb.Status = status;
      SysDclAst (&DclSysCommandAst, tkptr);
   }
   tkptr->QueuedSysCommand++;
}

/*****************************************************************************/
/*
A queued write to the subprocess "SYS$COMMAND" mailbox has completed.  This is 
the subprocesses' "SYS$COMMAND", supplying the DCL commands to execute.
*/ 

DclSysCommandAst (struct DclTaskStruct *tkptr)

{
   /*********/
   /* begin */
   /*********/

   if (Debug)
   {
      fprintf (stdout,
         "DclSysCommandAst() %d %d %08.08X %d %d %d %d %%X%08.08X\n",
         tkptr, tkptr->RequestPtr, tkptr->SubprocessPid,
         tkptr->QueuedSysCommand, tkptr->QueuedSysOutput,
         tkptr->QueuedCgiPlusIn, tkptr->QueuedHttpInput,
         tkptr->SysCommandIOsb.Status);
   }

   if (tkptr->QueuedSysCommand) tkptr->QueuedSysCommand--;

   if (tkptr->WatchItem &&
       (WatchEnabled & WATCH_DCL))
   {
      if (VMSnok (tkptr->SysCommandIOsb.Status))
         WatchThis (tkptr->RequestPtr, FI_LI, WATCH_DCL,
                    "WRITE SYS$COMMAND %X!8XL %!%M",
                    tkptr->SysCommandIOsb.Status,
                    tkptr->SysCommandIOsb.Status);
      else
         WatchThis (tkptr->RequestPtr, FI_LI, WATCH_DCL,
                    "WRITE SYS$COMMAND %X!8XL",
                    tkptr->SysCommandIOsb.Status);
   }

   /* if (effectively) no outstanding I/O then conclude the DCL task */
   if (tkptr->QueuedSysCommand <= tkptr->QueuedSysCommandAllowed &&
       !tkptr->QueuedSysOutput &&
       !tkptr->QueuedCgiPlusIn &&
       !tkptr->QueuedHttpInput)
   {
      DclConcludeTask (tkptr, false);
      return;
   }

   /* if I/O cancelled then just return */
   if (tkptr->SysCommandIOsb.Status == SS$_ABORT ||
       tkptr->SysCommandIOsb.Status == SS$_CANCEL)
      return;

   /* don't know why, first AST is delivered %x00000000, make %X00000001 */
   if (!tkptr->SysCommandIOsb.Status) tkptr->SysCommandIOsb.Status = 1;

   /* abort if an error writing SYS$COMMAND stream to subprocess */
   if (VMSnok (tkptr->SysCommandIOsb.Status))
   {
      tkptr->RequestPtr->rqResponse.ErrorTextPtr =
         MsgFor(tkptr->RequestPtr,MSG_SCRIPT_IPC);
      ErrorVmsStatus (tkptr->RequestPtr, tkptr->SysCommandIOsb.Status, FI_LI);
      DclConcludeTask (tkptr, true);
      return;
   }

   /* at least one DCL command was read indicating script activation */
   tkptr->SubprocessActivated = true;
}

/*****************************************************************************/
/*
Queue up a write of data to the subprocess "CGIPLUSIN" mailbox.  This is a
CGIplus subprocesses' CGI variable stream mailbox.             
*/ 

DclQioCgiPlusIn 
(
struct DclTaskStruct *tkptr,
char *DataPtr,
int DataLength
)
{
   int  status,
        DclAstStatus;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "DclQioCgiPlusIn() %d %d %d |%s|\n",
               tkptr, tkptr->RequestPtr, DataLength, DataPtr);

   if (DataPtr == NULL)
   {
      /* NULL pointer means write an end-of-file to the channel */

      if (tkptr->WatchItem &&
          (WatchEnabled & WATCH_DCL))
         WatchThis (tkptr->RequestPtr, FI_LI, WATCH_DCL,
                    "WRITE CGIPLUSIN EOF");

      status = sys$qio (0, tkptr->CgiPlusInChannel,
                        IO$_WRITEOF, &tkptr->CgiPlusInIOsb,
                        &DclCgiPlusInAst, tkptr,
                        0, 0, 0, 0, 0, 0);
   }
   else
   {
      if (tkptr->WatchItem &&
          (WatchEnabled & WATCH_DCL))
      {
         WatchThis (tkptr->RequestPtr, FI_LI, WATCH_DCL,
                    "WRITE CGIPLUSIN !UL bytes", DataLength);
         if (DataLength) WatchData (DataPtr, DataLength);
      }

      status = sys$qio (0, tkptr->CgiPlusInChannel,
                        IO$_WRITELBLK, &tkptr->CgiPlusInIOsb,
                        &DclCgiPlusInAst, tkptr,
                        DataPtr, DataLength, 0, 0, 0, 0);
   }

   if (VMSnok (status))
   {
      /* report this error via the AST */
      tkptr->CgiPlusInIOsb.Status = status;
      SysDclAst (&DclCgiPlusInAst, tkptr);
   }
   tkptr->QueuedCgiPlusIn++;
}

/*****************************************************************************/
/*
A queued write to the subprocess "CGIPLUSIN" mailbox has completed.  This is a
CGIplus subprocesses' CGI variable stream mailbox.             
*/ 

DclCgiPlusInAst (struct DclTaskStruct *tkptr)

{
   /*********/
   /* begin */
   /*********/

   if (Debug)
   {
      fprintf (stdout,
         "DclCgiPlusInAst() %d %d %08.08X %d %d %d %%X%08.08X\n",
         tkptr, tkptr->RequestPtr, tkptr->SubprocessPid, tkptr->QueuedCgiPlusIn,
         tkptr->QueuedSysOutput, tkptr->QueuedHttpInput,
         tkptr->CgiPlusInIOsb.Status);
   }

   if (tkptr->QueuedCgiPlusIn) tkptr->QueuedCgiPlusIn--;

   if (tkptr->WatchItem &&
       (WatchEnabled & WATCH_DCL))
   {
      if (VMSnok (tkptr->CgiPlusInIOsb.Status))
         WatchThis (tkptr->RequestPtr, FI_LI, WATCH_DCL,
                    "WRITE CGIPLUSIN %X!8XL %!%M",
                    tkptr->CgiPlusInIOsb.Status,
                    tkptr->CgiPlusInIOsb.Status);
      else
         WatchThis (tkptr->RequestPtr, FI_LI, WATCH_DCL,
                    "WRITE CGIPLUSIN %X!8XL",
                    tkptr->CgiPlusInIOsb.Status);
   }

   /* if (effectively) no outstanding I/O then conclude the DCL task */
   if (tkptr->QueuedSysCommand <= tkptr->QueuedSysCommandAllowed &&
       !tkptr->QueuedSysOutput &&
       !tkptr->QueuedCgiPlusIn &&
       !tkptr->QueuedHttpInput)
   {
      DclConcludeTask (tkptr, false);
      return;
   }

   /* if I/O cancelled then just return */
   if (tkptr->CgiPlusInIOsb.Status == SS$_ABORT ||
       tkptr->CgiPlusInIOsb.Status == SS$_CANCEL)
   {
      /* any 'cgiplusin' abort indicates the subprocess failed in some way */
      if (tkptr->CgiPlusInIOsb.Status == SS$_ABORT)
         tkptr->SubprocessActivated = false;
      return;
   }

   /* don't know why, first AST is delivered %x00000000, make %X00000001 */
   if (!tkptr->CgiPlusInIOsb.Status) tkptr->CgiPlusInIOsb.Status = 1;

   /* abort if an error writing CGIPLUSIN stream to subprocess */
   if (VMSnok (tkptr->CgiPlusInIOsb.Status))
   {
      tkptr->RequestPtr->rqResponse.ErrorTextPtr =
         MsgFor(tkptr->RequestPtr,MSG_SCRIPT_IPC);
      ErrorVmsStatus (tkptr->RequestPtr, tkptr->CgiPlusInIOsb.Status, FI_LI);
      DclConcludeTask (tkptr, true);
      return;
   }

   /* at least one CGIPLUSIN variable was read indicating script activation */
   tkptr->SubprocessActivated = true;
}

/*****************************************************************************/
/*
A queued write to the subprocess "HTTP$INPUT" mailbox has completed.
Provide more (possibly first) of the request body, or EOF.
*/ 

DclHttpInputAst (struct DclTaskStruct *tkptr)

{
   int  status,
        DclAstStatus,
        Length;
   char  *ContentPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
   {
      fprintf (stdout,
         "DclHttpInputAst() %d %d %08.08X %d %d %d %d %d %%X%08.08X\n",
         tkptr, tkptr->RequestPtr, tkptr->SubprocessPid,
         tkptr->QueuedSysCommand, tkptr->QueuedSysOutput,
         tkptr->QueuedCgiPlusIn, tkptr->QueuedHttpInput,
         tkptr->HttpInputIOsb.Count, tkptr->HttpInputIOsb.Status);
   }

   if (tkptr->QueuedHttpInput) tkptr->QueuedHttpInput--;

   if (tkptr->WatchItem &&
       (WatchEnabled & WATCH_DCL))
   {
      if (VMSnok (tkptr->HttpInputIOsb.Status))
         WatchThis (tkptr->RequestPtr, FI_LI, WATCH_DCL,
                    "WRITE HTTP$INPUT %X!8XL %!%M",
                    tkptr->HttpInputIOsb.Status,
                    tkptr->HttpInputIOsb.Status);
      else
         WatchThis (tkptr->RequestPtr, FI_LI, WATCH_DCL,
                    "WRITE HTTP$INPUT %X!8XL",
                    tkptr->HttpInputIOsb.Status);
   }

   /* if (effectively) no outstanding I/O then conclude the DCL task */
   if (tkptr->QueuedSysCommand <= tkptr->QueuedSysCommandAllowed &&
       !tkptr->QueuedSysOutput &&
       !tkptr->QueuedCgiPlusIn &&
       !tkptr->QueuedHttpInput)
   {
      DclConcludeTask (tkptr, false);
      return;
   }

   /* if I/O cancelled then just return */
   if (tkptr->HttpInputIOsb.Status == SS$_ABORT ||
       tkptr->HttpInputIOsb.Status == SS$_CANCEL)
      return;

   /* abort if an error writing HTTP stream to subprocess */
   if (VMSnok (tkptr->HttpInputIOsb.Status))
   {
      tkptr->RequestPtr->rqResponse.ErrorTextPtr =
         MsgFor(tkptr->RequestPtr,MSG_SCRIPT_IPC);
      ErrorVmsStatus (tkptr->RequestPtr, tkptr->HttpInputIOsb.Status, FI_LI);
      DclConcludeTask (tkptr, true);
      return;
   }

   if (Debug)
      fprintf (stdout, "%d %d\n", tkptr->ContentLength, tkptr->ContentPtr);

   if (tkptr->ContentLength)
   {
      if (tkptr->ContentLength > NetReadBufferSize)
         Length = NetReadBufferSize;
      else
         Length = tkptr->ContentLength;
      ContentPtr = tkptr->ContentPtr;
      tkptr->ContentPtr += Length;
      tkptr->ContentLength -= Length;

      if (tkptr->WatchItem &&
          (WatchEnabled & WATCH_DCL))
      {
         WatchThis (tkptr->RequestPtr, FI_LI, WATCH_DCL,
                    "WRITE SYS$INPUT !UL bytes", Length);
         if (Length) WatchDataDump (ContentPtr, Length);
      }

      status = sys$qio (0, tkptr->HttpInputChannel,
                        IO$_WRITELBLK, &tkptr->HttpInputIOsb,
                        &DclHttpInputAst, tkptr,
                        ContentPtr, Length, 0, 0, 0, 0);
   }
   else
   {
      if (tkptr->WatchItem &&
          (WatchEnabled & WATCH_DCL))
         WatchThis (tkptr->RequestPtr, FI_LI, WATCH_DCL,
                    "WRITE SYS$INPUT EOF");

      status = sys$qio (0, tkptr->HttpInputChannel,
                        IO$_WRITEOF, &tkptr->HttpInputIOsb,
                        &DclHttpInputAst, tkptr,
                        0, 0, 0, 0, 0, 0);
   }

   if (VMSnok (status))
   {
      /* report this error via the AST */
      tkptr->HttpInputIOsb.Status = status;
      SysDclAst (&DclHttpInputAst, tkptr);
   }
   tkptr->QueuedHttpInput++;
}

/*****************************************************************************/
/*
Send DCL commands to the CGI script or DCL command subprocess' SYS$COMMAND.
This sets up the DCL environment (defines logical names, assigns symbols)
executes the procedure or image.
*/ 

int DclCgiScriptSysCommand (struct DclTaskStruct *tkptr)

{
   static char  DclHttpdVerify [] = "!'F$VERIFY(F$TRNLNM(\"HTTPD$VERIFY\"))";
   static char  DeleteIsDelete [] = "DELETE=\"DELETE\"";
   static char  DeleteAllLocalSymbol[] = "DELETE/SYMBOL/LOCAL/ALL";
   static char  DeleteAllGlobalSymbol[] = "DELETE/SYMBOL/GLOBAL/ALL";
   static char  HttpdLogin [] =
                "IF F$TRNLNM(\"HTTPD$LOGIN\").NES.\"\" THEN @HTTPD$LOGIN";
   static char  NoVerify [] = "!'F$VERIFY(0)";
   static char  SetNoOn[] = "SET NOON";
   /* this symbol is required by CGI.C module for single quote substitution */
   static char  StopId [] = "STOP/ID=0";
   static char  WriteIsWrite [] = "WRITE=\"WRITE\"";

   static $DESCRIPTOR (WriteDclEofFaoDsc, "WRITE SYS$OUTPUT \"!AZ\"");

   static $DESCRIPTOR (DefineHttpInputFaoDsc,
                       "DEFINE/NOLOG/SUPER HTTP$INPUT !AZ");
   static $DESCRIPTOR (DefineSysInputFaoDsc,
                       "DEFINE/NOLOG/SUPER SYS$INPUT !AZ");

#if CGIPLUS_CALLOUT_FOR_CGI

   static $DESCRIPTOR (DefineCgiPlusInFaoDsc,
                       "DEFINE/NOLOG/SUPER CGIPLUSIN !AZ");
   static $DESCRIPTOR (DefineCgiPlusEotFaoDsc,
                       "DEFINE/NOLOG/SUPER CGIPLUSEOT \"!AZ\"");
   static $DESCRIPTOR (DefineCgiPlusEscFaoDsc,
                       "DEFINE/NOLOG/SUPER CGIPLUSESC \"!AZ\"");

#endif /* CGIPLUS_CALLOUT_FOR_CGI */

#if STREAMS_FOR_APACHE

   static $DESCRIPTOR (DefineApacheInputFaoDsc,
                       "DEFINE/NOLOG/SUPER APACHE_INPUT !AZ");

#endif /* STREAMS_FOR_APACHE */

#if STREAMS_FOR_PURVEYOR_AND_CERN

   static $DESCRIPTOR (DefineWwwInFaoDsc,
                       "DEFINE/NOLOG/SUPER WWW_IN !AZ");
   static char  DefineWwwOut[] = "DEFINE/NOLOG/SUPER WWW_OUT SYS$OUTPUT:";

#endif /* STREAMS_FOR_PURVEYOR_AND_CERN */

   register struct RequestStruct *rqptr;
   register char  c;
   register char  *cptr, *sptr, *zptr;

   int  status,
        Count;
   unsigned short  Length;
   char  FormFieldName [256],
         Scratch [1024],
         DclLine [256];
   $DESCRIPTOR (DclLineDsc, DclLine);

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "DclCgiScriptSysCommand() |%s|%s|%s|\n",
               tkptr->DclCommand, tkptr->ScriptFileName,
               tkptr->ScriptRunTimePtr);

   /* get the pointer to the request structure */
   rqptr = tkptr->RequestPtr;

   DclQioSysCommand (tkptr, DclHttpdVerify, sizeof(DclHttpdVerify)-1);
   DclQioSysCommand (tkptr, SetNoOn, sizeof(SetNoOn)-1);
   DclQioSysCommand (tkptr, DeleteIsDelete, sizeof(DeleteIsDelete)-1);
   DclQioSysCommand (tkptr, DeleteAllLocalSymbol,
                     sizeof(DeleteAllLocalSymbol)-1);
   DclQioSysCommand (tkptr, DeleteAllGlobalSymbol,
                     sizeof(DeleteAllGlobalSymbol)-1);

   if (tkptr->TaskType == DCL_TASK_TYPE_CGI_SCRIPT)
   {
      sys$fao (&DefineSysInputFaoDsc, &Length, &DclLineDsc,
               tkptr->HttpInputDevName);
      DclLine[Length] = '\0';
      DclQioSysCommand (tkptr, DclLine, Length);

      /* for backward compatibility */
      sys$fao (&DefineHttpInputFaoDsc, &Length, &DclLineDsc,
               tkptr->HttpInputDevName);
      DclLine[Length] = '\0';
      DclQioSysCommand (tkptr, DclLine, Length);

#if STREAMS_FOR_PURVEYOR_AND_CERN

      /* for Purveyor, Cern backward compatibility */
      sys$fao (&DefineWwwInFaoDsc, &Length, &DclLineDsc,
               tkptr->HttpInputDevName);
      DclLine[Length] = '\0';
      DclQioSysCommand (tkptr, DclLine, Length);
      DclQioSysCommand (tkptr, DefineWwwOut, sizeof(DefineWwwOut)-1);

#endif /* STREAMS_FOR_PURVEYOR_AND_CERN */

#if STREAMS_FOR_APACHE

      /* for VMS Apache forward compatibility :^) */
      sys$fao (&DefineApacheInputFaoDsc, &Length, &DclLineDsc,
               tkptr->HttpInputDevName);
      DclLine[Length] = '\0';
      DclQioSysCommand (tkptr, DclLine, Length);

#endif /* STREAMS_FOR_APACHE */

#if CGIPLUS_CALLOUT_FOR_CGI

      sys$fao (&DefineCgiPlusInFaoDsc, &Length, &DclLineDsc,
               tkptr->CgiPlusInDevName);
      DclLine[Length] = '\0';
      status = DclQioSysCommand (tkptr, DclLine, Length);

      sys$fao (&DefineCgiPlusEotFaoDsc, &Length, &DclLineDsc,
               tkptr->CgiEot);
      DclLine[Length] = '\0';
      status = DclQioSysCommand (tkptr, DclLine, Length);

      sys$fao (&DefineCgiPlusEscFaoDsc, &Length, &DclLineDsc,
               tkptr->CgiEsc);
      DclLine[Length] = '\0';
      status = DclQioSysCommand (tkptr, DclLine, Length);

#endif /* CGIPLUS_CALLOUT_FOR_CGI */
   }

   /*****************/
   /* CGI variables */
   /*****************/

   /*
      CGI script requests can only call this once.
      DCL commands from SSI documents can call this multiple times.
      Once the CGI information is generated is is essentially static
      for the life of the request ... so only generate it the once!
   */
   if (rqptr->rqCgi.BufferPtr == NULL)
      if (VMSnok (status = CgiGenerateVariables (rqptr, CGI_VARIABLE_DCL)))
         return (status);

   cptr = rqptr->rqCgi.BufferPtr;
   for (;;)
   {
      if (!(Length = *(short*)cptr)) break;
      DclQioSysCommand (tkptr, cptr+sizeof(short), Length-1);
      cptr += Length + sizeof(short);
   }

   /*******************************/
   /* DCL command/procedure/image */
   /*******************************/

   DclQioSysCommand (tkptr, HttpdLogin, sizeof(HttpdLogin)-1);

   if (tkptr->ScriptFileName[0])
   {
      sptr = DclLine;
      cptr = tkptr->ScriptRunTimePtr;
      if (*(unsigned short*)cptr == '@\0')
      {
         /* DCL procedure */
         while (*cptr) *sptr++ = *cptr++;
      }
      else
      if (*(unsigned short*)cptr == '$\0')
      {
         /* execute an image */
         cptr = "RUN ";
         while (*cptr) *sptr++ = *cptr++;
      }
      else
      if (*cptr == '@' || *cptr == '$')
      {
         /* foreign-verb DCL procedure or executable, create the verb */
         cptr++;
         while (*cptr) *sptr++ = *cptr++;
         *sptr = '\0';
         DclQioSysCommand (tkptr, DclLine, sptr - DclLine);

         /* now place it as the verb before the script file */
         sptr = DclLine;
         cptr = tkptr->ScriptRunTimePtr + 1;
         while (*cptr && *cptr != '=') *sptr++ = *cptr++;
         *sptr++ = ' ';
      }
      else
      {
         /* verb must already exist on site, place before the script file */
         while (*cptr) *sptr++ = *cptr++;
         *sptr++ = ' ';
      }

      cptr = tkptr->SearchOds.ResFileName;
      while (*cptr) *sptr++ = *cptr++;
      *sptr = '\0';
      Length = sptr - (cptr = DclLine);
      DclQioSysCommand (tkptr, cptr, Length);
   }
   else
   {
      /* multiple commands may be separated by newlines */
      sptr = tkptr->DclCommand;
      while (*sptr)
      {
         cptr = sptr;
         while (*sptr && *sptr != '\n') sptr++;
         Length = sptr - cptr;
         if (*sptr == '\n') *sptr++ = '\0';
         DclQioSysCommand (tkptr, cptr, Length);
      }
   }

   /*********************/
   /* after script runs */
   /*********************/

   if (DclUseZombies)
   {
      DclQioSysCommand (tkptr, NoVerify, sizeof(NoVerify)-1);
      DclQioSysCommand (tkptr, WriteIsWrite, sizeof(WriteIsWrite)-1);

      sys$fao (&WriteDclEofFaoDsc, &Length, &DclLineDsc, tkptr->CgiEof);
      DclLine[Length] = '\0';
      DclQioSysCommand (tkptr, DclLine, Length);

      /* do not send an end-of-file! */
      return (SS$_NORMAL);
   }
   else
   {
      /* ensure subprocess terminates! */
      DclQioSysCommand (tkptr, StopId, sizeof(StopId)-1);

      /* send end-of-file */
      DclQioSysCommand (tkptr, NULL, 0);
   }

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
Send DCL commands to the CGIplus subprocess' SYS$COMMAND. This sets up the DCL
environment (defines logical names, assigns symbols) executes the procedure or
image.
*/ 

DclCgiPlusScriptSysCommand (struct DclTaskStruct *tkptr)

{
   static char  DclHttpdVerify [] = "!'F$VERIFY(F$TRNLNM(\"HTTPD$VERIFY\"))";
   static char  DeleteIsDelete [] = "DELETE=\"DELETE\"";
   static char  DeleteAllLocalSymbol[] = "DELETE/SYMBOL/LOCAL/ALL";
   static char  DeleteAllGlobalSymbol[] = "DELETE/SYMBOL/GLOBAL/ALL";
   static char  HttpdLogin [] =
                "IF F$TRNLNM(\"HTTPD$LOGIN\").NES.\"\" THEN @HTTPD$LOGIN";
   static char  SetNoOn [] = "SET NOON";
   static char  StopId [] = "STOP/ID=0";

   static $DESCRIPTOR (DefineCgiPlusInFaoDsc,
                       "DEFINE/NOLOG/SUPER CGIPLUSIN !AZ");
   static $DESCRIPTOR (DefineCgiPlusEofFaoDsc,
                       "DEFINE/NOLOG/SUPER CGIPLUSEOF \"!AZ\"");
   static $DESCRIPTOR (DefineCgiPlusEotFaoDsc,
                       "DEFINE/NOLOG/SUPER CGIPLUSEOT \"!AZ\"");
   static $DESCRIPTOR (DefineCgiPlusEscFaoDsc,
                       "DEFINE/NOLOG/SUPER CGIPLUSESC \"!AZ\"");
   static $DESCRIPTOR (DefineHttpInputFaoDsc,
                       "DEFINE/NOLOG/SUPER HTTP$INPUT !AZ");
   static $DESCRIPTOR (DefineSysInputFaoDsc,
                       "DEFINE/NOLOG/SUPER SYS$INPUT !AZ");

   register char  *cptr, *sptr;
   register struct RequestStruct *rqptr;

   int  status;
   unsigned short  Length;
   char  DclLine [256+64];
   $DESCRIPTOR (DclLineDsc, DclLine);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DclCgiPlusScriptSysCommand()\n");

   /* get the pointer to the request structure */
   rqptr = tkptr->RequestPtr;

   DclQioSysCommand (tkptr, DclHttpdVerify, sizeof(DclHttpdVerify)-1);
   DclQioSysCommand (tkptr, SetNoOn, sizeof(SetNoOn)-1);
   DclQioSysCommand (tkptr, DeleteIsDelete, sizeof(DeleteIsDelete)-1);
   DclQioSysCommand (tkptr, DeleteAllLocalSymbol,
                     sizeof(DeleteAllLocalSymbol)-1);
   DclQioSysCommand (tkptr, DeleteAllGlobalSymbol,
                     sizeof(DeleteAllGlobalSymbol)-1);

   sys$fao (&DefineSysInputFaoDsc, &Length, &DclLineDsc,
            tkptr->HttpInputDevName);
   DclLine[Length] = '\0';
   DclQioSysCommand (tkptr, DclLine, Length);

   /* for backward compatibility */
   sys$fao (&DefineHttpInputFaoDsc, &Length, &DclLineDsc,
            tkptr->HttpInputDevName);
   DclLine[Length] = '\0';
   DclQioSysCommand (tkptr, DclLine, Length);

   sys$fao (&DefineCgiPlusInFaoDsc, &Length, &DclLineDsc,
            tkptr->CgiPlusInDevName);
   DclLine[Length] = '\0';
   status = DclQioSysCommand (tkptr, DclLine, Length);

   sys$fao (&DefineCgiPlusEofFaoDsc, &Length, &DclLineDsc,
            tkptr->CgiEof);
   DclLine[Length] = '\0';
   status = DclQioSysCommand (tkptr, DclLine, Length);

   sys$fao (&DefineCgiPlusEotFaoDsc, &Length, &DclLineDsc,
            tkptr->CgiEot);
   DclLine[Length] = '\0';
   status = DclQioSysCommand (tkptr, DclLine, Length);

   sys$fao (&DefineCgiPlusEscFaoDsc, &Length, &DclLineDsc,
            tkptr->CgiEsc);
   DclLine[Length] = '\0';
   status = DclQioSysCommand (tkptr, DclLine, Length);

   /***********************/
   /* DCL procedure/image */
   /***********************/

   DclQioSysCommand (tkptr, HttpdLogin, sizeof(HttpdLogin)-1);

   sptr = DclLine;

   if (tkptr->TaskType == DCL_TASK_TYPE_CGIPLUS_SCRIPT)
   {
      cptr = tkptr->ScriptRunTimePtr;
      if (*(unsigned short*)cptr == '@\0')
      {
         /* DCL procedure */
         while (*cptr) *sptr++ = *cptr++;
      }
      else
      if (*(unsigned short*)cptr == '$\0')
      {
         /* execute an image */
         cptr = "RUN ";
         while (*cptr) *sptr++ = *cptr++;
      }
      else
      if (*cptr == '@' || *cptr == '$')
      {
         /* foreign-verb DCL procedure or executable, create the verb */
         cptr++;
         while (*cptr) *sptr++ = *cptr++;
         *sptr = '\0';
         DclQioSysCommand (tkptr, DclLine, sptr - DclLine);

         /* now place it as the verb before the script file */
         sptr = DclLine;
         cptr = tkptr->ScriptRunTimePtr + 1;
         while (*cptr && *cptr != '=') *sptr++ = *cptr++;
         *sptr++ = ' ';
         *sptr = '\0';
      }
      else
      {
         /* verb must already exist on site, place before the script file */
         while (*cptr) *sptr++ = *cptr++;
         *sptr++ = ' ';
         *sptr = '\0';
      }

      cptr = tkptr->SearchOds.ResFileName;
      while (*cptr) *sptr++ = *cptr++;
   }

   *sptr = '\0';
   Length = sptr - (cptr = DclLine);
   if (Debug) fprintf (stdout, "|%s|\n", DclLine);
   DclQioSysCommand (tkptr, cptr, Length);

   /* ensure subprocess terminates! */
   DclQioSysCommand (tkptr, StopId, sizeof(StopId)-1);

   /* send end-of-file */
   DclQioSysCommand (tkptr, NULL, 0);
}

/*****************************************************************************/
/*
Send CGI variables to the subprocess' CGIPLUSIN input stream.
*/ 

int DclCgiPlusScriptCgiPlusIn (struct DclTaskStruct *tkptr)

{
   register char  *cptr;
   register struct RequestStruct *rqptr;

   int  status;
   unsigned short  Length;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DclCgiPlusScriptCgiPlusIn()\n");

   /* get the pointer to the request structure */
   rqptr = tkptr->RequestPtr;

   /* just a line for "start-of-request" that can always be discarded */
   DclQioCgiPlusIn (tkptr, "!", 1);

   if (VMSnok (status = CgiGenerateVariables (rqptr, CGI_VARIABLE_STREAM)))
      return (status);

   cptr = rqptr->rqCgi.BufferPtr;
   for (;;)
   {
      if (!(Length = *(short*)cptr)) break;
      DclQioCgiPlusIn (tkptr, cptr+sizeof(short), Length-1);
      cptr += Length + sizeof(short);
   }

   /* empty line terminates CGI variables */
   DclQioCgiPlusIn (tkptr, "", 0);

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
This function is available for an agent callout to write output to the agent's
CGIplus input mailbox.  It may be called one or more times from a single
callout (the only thing to be mindful of is BYTLM and the capacity of the
CGIPLUSIN mailbox.
*/ 

DclCgiPlusCalloutQio
(
struct RequestStruct *rqptr,
char *DataPtr,
int DataLength
)
{
   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "DclCgiPlusCalloutQio() %d %d %d |%s|\n",
               rqptr, rqptr->DclTaskPtr, DataLength, DataPtr);

   if (rqptr->DclTaskPtr == NULL)
      ErrorExitVmsStatus (SS$_BUGCHECK, ErrorSanityCheck, FI_LI);

   DclQioCgiPlusIn (rqptr->DclTaskPtr, DataPtr, DataLength);
}

/*****************************************************************************/
/*
Default callout for CGIplus scripting.  In other words; ask the the server to
do something for a CGIplus script.  See prologue to this module for description
of currently support functionlity.
*/ 

DclCgiPlusDefaultCallout (struct RequestStruct *rqptr)

{
   static char  ResponseBadParam [] = "400 Bad parameter",
                ResponseUnauthorized [] = "401 Unauthorized",
                ResponseForbidden [] = "403 Forbidden",
                ResponseSuccess [] = "200 Success",
                ResponseUnknown [] = "400 Unknown request";

   register char  *cptr;

   int  status,
        CgiPlusLifeTime;
   char  *FileNamePtr,
         *OutputPtr;
   char  Scratch [1024+4];

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "DclCgiPlusDefaultCallout() %d |%s|\n",
               rqptr->rqCgi.CalloutOutputCount,
               rqptr->rqCgi.CalloutOutputPtr);

   if ((OutputPtr = rqptr->rqCgi.CalloutOutputPtr) == NULL &&
       (rqptr->rqCgi.CalloutOutputCount == 0 ||
        rqptr->rqCgi.CalloutOutputCount == -1))
   {
      /* indicates the script has sent a begin or end escape sequence */
      return;
   }

   if (toupper(OutputPtr[0]) == 'A' &&
       strsame (OutputPtr, "AUTH-FILE:", 10))
   {
      /* authorize access to a particular file */
      for (cptr = OutputPtr+10; *cptr && isspace(*cptr); cptr++);
      if (!*cptr)
      {
         DclCgiPlusCalloutQio (rqptr, ResponseBadParam,
                                      sizeof(ResponseBadParam)-1);
         return;
      }
      FileNamePtr = cptr;
      while (*cptr && !isspace(*cptr)) cptr++;
      if (rqptr->rqAuth.VmsUserProfileLength)
      {
         status = AuthVmsCheckUserAccess (rqptr, FileNamePtr, cptr-FileNamePtr);
         if (status == RMS$_PRV)
            DclCgiPlusCalloutQio (rqptr, ResponseForbidden,
                                         sizeof(ResponseForbidden)-1);
         else
         if (VMSok (status))
            DclCgiPlusCalloutQio (rqptr, ResponseSuccess,
                                         sizeof(ResponseSuccess)-1);
         else
         {
            /* error reported by access check */
            WriteFao (Scratch, sizeof(Scratch), NULL, "400 %!%M", status);
            DclCgiPlusCalloutQio (rqptr, Scratch, strlen(Scratch));
         }
      }
      else
      {
         /* i.e. request does not have a VMS profile associated */
         DclCgiPlusCalloutQio (rqptr, ResponseUnauthorized,
                                      sizeof(ResponseUnauthorized)-1);
      }
      return;
   }

   if (toupper(OutputPtr[0]) == 'L' &&
       strsame (OutputPtr, "LIFETIME:", 9))
   {
      /* let a CGIplus script set/reset its own lifetime */
      for (cptr = OutputPtr+9; *cptr && isspace(*cptr); cptr++);
      if (isdigit(*cptr))
      {
         /* zero makes the script immune to supervisor purging */
         if ((CgiPlusLifeTime = atoi(cptr)) == 0) CgiPlusLifeTime = -1;
         rqptr->DclTaskPtr->LifeTimeCount = CgiPlusLifeTime;
         DclCgiPlusCalloutQio (rqptr, ResponseSuccess,
                                      sizeof(ResponseSuccess)-1);
      }               
      else
         DclCgiPlusCalloutQio (rqptr, ResponseBadParam,
                                      sizeof(ResponseBadParam)-1);
      return;
   }

   if (toupper(OutputPtr[0]) == 'M' &&
       strsame (OutputPtr, "MAP-FILE:", 9))
   {
      Scratch[4] = '\0';
      for (cptr = OutputPtr+9; *cptr && isspace(*cptr); cptr++);
      if (!*cptr)
      {
         DclCgiPlusCalloutQio (rqptr, ResponseBadParam,
                                      sizeof(ResponseBadParam)-1);
         return;
      }
      cptr = MapUrl_Map (Scratch+4, sizeof(Scratch)-4, cptr, 0,
                         NULL, 0, NULL, 0, NULL, rqptr);
      if (!cptr[0] && cptr[1])
      {
         memcpy (Scratch, "400 ", 4);
         strcpy (Scratch+4, cptr+1);
         DclCgiPlusCalloutQio (rqptr, Scratch, strlen(Scratch));
      }
      else
      {
         memcpy (Scratch, "200 ", 4);
         DclCgiPlusCalloutQio (rqptr, Scratch, strlen(Scratch));
      }
      return;
   }

   if (toupper(OutputPtr[0]) == 'M' &&
       strsame (OutputPtr, "MAP-PATH:", 9))
   {
      Scratch[4] = '\0';
      for (cptr = OutputPtr+9; *cptr && isspace(*cptr); cptr++);
      if (!*cptr)
      {
         DclCgiPlusCalloutQio (rqptr, ResponseBadParam,
                                      sizeof(ResponseBadParam)-1);
         return;
      }
      cptr = MapUrl_Map (cptr, 0, Scratch+4, sizeof(Scratch)-4,
                         NULL, 0, NULL, 0, NULL, rqptr);
      if (!cptr[0] && cptr[1])
      {
         memcpy (Scratch, "400 ", 4);
         strcpy (Scratch+4, cptr+1);
         DclCgiPlusCalloutQio (rqptr, Scratch, strlen(Scratch));
      }
      else
      {
         memcpy (Scratch, "200 ", 4);
         DclCgiPlusCalloutQio (rqptr, Scratch, strlen(Scratch));
      }
      return;
   }

   if (toupper(OutputPtr[0]) == 'N' &&
       strsame (OutputPtr, "NOOP:", 5))
   {
      /* could be used for WATCHable debugging information, comments, etc. */
      DclCgiPlusCalloutQio (rqptr, ResponseSuccess, sizeof(ResponseSuccess)-1);
      return;
   }

   DclCgiPlusCalloutQio (rqptr, ResponseUnknown, sizeof(ResponseUnknown)-1);
}

/*****************************************************************************/
/*
Every minute scan the list of DCL subprocesses looking for those whose
lifetimes have expired. Run those subprocesses down! As this is a one minute
tick set the lifetime counters to configuration plus one to ensure at least
that number of minutes before expiry.  A lifetime count of -1 indicates the
subprocess has requested that it be immune to supervisor purging.
*/

DclSupervisor (boolean TimerExpired)

{
   static boolean  TimerSet = false;
   static unsigned long  OneMinuteDelta [2] = { -600000000, -1 };

   register struct ListEntryStruct  *leptr;
   register struct DclTaskStruct  *tkptr;

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DclSupervisor() %d\n", TimerExpired);

   if (!TimerExpired)
   {
      /* new request, if the supervisor is already running just return */
      if (TimerSet) return;

      /* kick the supervisor timer into life, and then return */
      if (VMSnok (status =
          sys$setimr (0, &OneMinuteDelta, &DclSupervisor, true, 0)))
         ErrorExitVmsStatus (status, "sys$setimr()", FI_LI);
      TimerSet = true;
      return;
   }

   /* must have got here via the supervisor's timer expiring */
   TimerSet = false;

   for (leptr = DclTaskList.HeadPtr; leptr != NULL; leptr = leptr->NextPtr)
   {
      tkptr = (struct DclTaskStruct*)leptr;
      if (Debug) DclTaskItemDebug (leptr, tkptr);
 
      /* let the output timer take care of those in use, ignore marked */
      if (tkptr->RequestPtr != NULL ||
          tkptr->MarkedForDelete ||
          tkptr->LifeTimeCount == -1) continue;

      if (tkptr->LifeTimeCount > 0)
      {
         if (--tkptr->LifeTimeCount)
         {
            TimerSet = true;
            continue;
         }
      }

      /* process timer has expired, exterminate ... exxterrminnaattte */
      tkptr->MarkedForDelete = true;
      DclConcludeTask (tkptr, true);
      if (tkptr->TaskType == DCL_TASK_TYPE_CGIPLUS_SCRIPT)
         DclCgiPlusLifeTimePurgeCount++;
      else
         DclZombieLifeTimePurgeCount++;
   }

   if (TimerSet)
   {
      /* at least one item in the connect list is still counting down */
      if (VMSnok (status =
          sys$setimr (0, &OneMinuteDelta, &DclSupervisor, true, 0)))
         ErrorExitVmsStatus (status, "sys$setimr()", FI_LI);
   }
   else
   {
      /* purge the script name cache when there are no more zombies/CGIplus */
      DclPurgeScriptNameCache ();
   }
}

/*****************************************************************************/
/*
************
*** NOTE ***  This function takes a pointer to a request!!!
************  Due to it being a general report processing function.

Return a report on the DCL task structure.  This function blocks while
executing.
*/ 

DclScriptingReport
(
struct RequestStruct *rqptr,
void *NextTaskFunction
)
{
   static char  ResponseFao [] =
"<HTML>\n\
<HEAD>\n\
!AZ\
<TITLE>HTTPd !AZ ... CGI/DCL Scripting</TITLE>\n\
</HEAD>\n\
!AZ\n\
<H2><NOBR>HTTPd !AZ</NOBR></H2>\n\
<H3>CGI/DCL Scripting</H3>\n\
!20%W\n\
\
<P><TABLE CELLPADDING=0 CELLSPACING=0 BORDER=0>\n\
<TR><TD VALIGN=top>\n\
\
<TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Statistics</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TH ALIGN=right>CGI-script:</TH><TD ALIGN=left>!UL</TD></TR>\n\
<TH ALIGN=right>CGIplus-script:</TH>\
<TD ALIGN=left>!UL / !UL &nbsp;<I>(all/reused)</I></TD></TR>\n\
<TH ALIGN=right>Autoscripted:</TH><TD ALIGN=left>!UL</TD></TR>\n\
<TH ALIGN=right>CLI:</TH><TD ALIGN=left>!UL</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
</TD><TD>&nbsp;&nbsp;&nbsp;&nbsp;</TD><TD VALIGN=top>\n\
\
<TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Subprocesses</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TH ALIGN=right COLSPAN=2>Spawned:</TH><TD ALIGN=left>!UL</TD></TR>\n\
<TH ALIGN=right COLSPAN=2>Current:</TH><TD ALIGN=left>!UL</TD></TR>\n\
<TH ALIGN=left><U>Soft-Limit</U></TH>\
<TH ALIGN=right>Value:</TH><TD ALIGN=left>!UL</TD></TR>\n\
<TH ALIGN=right COLSPAN=2>Purged-at:</TH><TD ALIGN=left>!UL</TD></TR>\n\
<TH ALIGN=left><U>Hard-Limit</U></TH>\
<TH ALIGN=right>Value:</TH><TD ALIGN=left>!UL</TD></TR>\n\
<TH ALIGN=right COLSPAN=2>Purged-at:</TH><TD ALIGN=left>!UL</TD></TR>\n\
<TH ALIGN=left><U>Purged</U></TH>\
<TH ALIGN=right>Soft-Limit:</TH><TD ALIGN=left>!UL</TD></TR>\n\
<TH ALIGN=right COLSPAN=2>Explicit:</TH><TD ALIGN=left>!UL</TD></TR>\n\
<TH ALIGN=left><U>Zombie</U></TH>\
<TH ALIGN=right>Lifetime:</TH><TD ALIGN=left>!%%</TD></TR>\n\
<TH ALIGN=right COLSPAN=2>Purged-at:</TH><TD ALIGN=left>!UL</TD></TR>\n\
<TH ALIGN=left><U>CGIplus</U></TH>\
<TH ALIGN=right>Lifetime:</TH><TD ALIGN=left>!%%</TD></TR>\n\
<TH ALIGN=right COLSPAN=2>Purged-at:</TH><TD ALIGN=left>!UL</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH COLSPAN=8>DCL Task List</TH></TR>\n\
<TR><TH></TH><TH>Script/DCL</TH><TH>PID</TH><TH>Lifetime</TH>\
<TH>Zombie</TH><TH>CGIplus</TH><TH>Total</TH><TH>Last</TH></TR>\n\
<TR><TH></TH><TH COLSPAN=1>Client</TH><TH COLSPAN=6>Request</TH></TR>\n\
<TR><TH COLSPAN=8></TH></TR>\n";

   static char  TaskFao [] =
"<TR><TD ALIGN=right>!UL</TD>\
<TD ALIGN=left>!AZ</TD><TD>!%%</TD><TD ALIGN=right>!AZ!UL!AZ</TD>\
<TD ALIGN=right>!%%</TD><TD ALIGN=right>!%%</TD><TD ALIGN=right>!UL</TD>\
<TD><FONT SIZE=-2>!23%W</FONT></TD></TR>\n\
!%%";

   static char  ScriptNameCacheFao [] =
"</TABLE>\n\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH COLSPAN=6>Script Name Cache</TH></TR>\n\
<TR><TH></TH><TH>Name</TH><TH>File</TH><TH>Hits</TH>\
<TH>Total</TH><TH>Last</TH></TR>\n\
<TR><TH COLSPAN=4></TH></TR>\n";

   static char  NameCacheFao [] =
"<TR><TD ALIGN=right>!UL</TD><TD ALIGN=left>!AZ</TD>\
<TD ALIGN=left><TT>!AZ</TT></TD><TD ALIGN=right>!UL</TD>\
<TD ALIGN=right>!UL</TD>\
<TD ALIGN=left><FONT SIZE=-2>!23%W</FONT></TD></TR>\n";

   static char  ButtonsFao [] =
"</TABLE>\n\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=0>\n\
<TR><TD>\n\
<FORM METHOD=GET ACTION=\"!AZ\">\n\
<INPUT TYPE=submit VALUE=\" Purge \">\n\
</FORM>\n\
</TD><TD>&nbsp;</TD><TD>\n\
<FORM METHOD=GET ACTION=\"!AZ\">\n\
<INPUT TYPE=submit VALUE=\" Force Delete \">\n\
</FORM>\n\
</TD></TR>\n\
</TABLE>\n\
</BODY>\n\
</HTML>\n";

   register unsigned long  *vecptr;
   register struct DclScriptNameCacheStruct  *captr;
   register struct DclTaskStruct  *tkptr;
   register struct ListEntryStruct  *leptr;

   int  status,
        Count;
   unsigned long  FaoVector [64];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DclScriptingReport()\n");

   rqptr->rqResponse.PreExpired = PRE_EXPIRE_ADMIN;
   HTTP_HEADER_200_HTML (rqptr);

   vecptr = FaoVector;

   *vecptr++ = HtmlMetaInfo (rqptr, NULL);
   *vecptr++ = ServerHostPort;
   *vecptr++ = Config.cfServer.AdminBodyTag;
   *vecptr++ = ServerHostPort;
   *vecptr++ = &rqptr->rqTime.Vms64bit;

   *vecptr++ = Accounting.DoScriptCount;
   *vecptr++ = Accounting.DoCgiPlusScriptCount;
   *vecptr++ = Accounting.DclCgiPlusReusedCount;
   *vecptr++ = Accounting.DoAutoScriptCount;
   *vecptr++ = Accounting.DoDclCommandCount;

   *vecptr++ = Accounting.DclSpawnCount;
   *vecptr++ = DclSubprocessCount;
   *vecptr++ = DclSubprocessSoftLimit;
   *vecptr++ = DclSoftLimitPurgeCount;
   *vecptr++ = DclSubprocessHardLimit;
   *vecptr++ = DclHitHardLimitCount;
   *vecptr++ = DclPurgeCount;
   *vecptr++ = DclPurgeAllSubprocessesCount;
   if (DclUseZombies && Config.cfScript.ZombieLifeTime)
   {
      *vecptr++ = "!UL";
      *vecptr++ = Config.cfScript.ZombieLifeTime;
   }
   else
      *vecptr++ = "[disabled]";
   *vecptr++ = DclZombieLifeTimePurgeCount;
   if (Config.cfScript.CgiPlusLifeTime)
   {
      *vecptr++ = "!UL";
      *vecptr++ = Config.cfScript.CgiPlusLifeTime;
   }
   else
      *vecptr++ = "<I>none</I>";
   *vecptr++ = DclCgiPlusLifeTimePurgeCount;

   status = NetWriteFaol (rqptr, ResponseFao, &FaoVector);
   if (VMSnok (status)) ErrorNoticed (status, "NetWriteFaol()", FI_LI);

   /* task list entries */
   Count = 0;
   for (leptr = DclTaskList.HeadPtr; leptr != NULL; leptr = leptr->NextPtr)
   {
      tkptr = (struct DclTaskStruct*)leptr;

      vecptr = FaoVector;

      *vecptr++ = ++Count;

      if (tkptr->TaskType == DCL_TASK_TYPE_CLI)
         *vecptr++ = tkptr->DclCommand;
      else
         *vecptr++ = tkptr->ScriptName;

      if (tkptr->SubprocessPid)
      {
         *vecptr++ = "<A HREF=\"!AZ?pid=!8XL\"><TT>!8XL</TT></A>";
         *vecptr++ = INTERNAL_REPORT_SHOW_PROCESS;
         *vecptr++ = tkptr->SubprocessPid;
         *vecptr++ = tkptr->SubprocessPid;
      }
      else
         *vecptr++ = "";

      if (tkptr->LifeTimeCount)
         *vecptr++ = "<B>";
      else
         *vecptr++ = "";
      /* fudge allows for correct reporting of lifetime */
      if ((tkptr->TaskType == DCL_TASK_TYPE_CGIPLUS_SCRIPT &&
           tkptr->LifeTimeCount == Config.cfScript.CgiPlusLifeTime+1) ||
          (tkptr->TaskType != DCL_TASK_TYPE_CGIPLUS_SCRIPT &&
           tkptr->LifeTimeCount == Config.cfScript.ZombieLifeTime+1))
         *vecptr++ = tkptr->LifeTimeCount-1;
      else
         *vecptr++ = tkptr->LifeTimeCount;
      if (tkptr->LifeTimeCount)
         *vecptr++ = "</B>";
      else
         *vecptr++ = "";

      if (tkptr->ZombieCount)
      {
         *vecptr++ = "!UL";
         *vecptr++ = tkptr->ZombieCount;
      }
      else
         *vecptr++ = "";

      if (tkptr->TaskType == DCL_TASK_TYPE_CGIPLUS_SCRIPT)
      {
         *vecptr++ = "!UL";
         *vecptr++ = tkptr->CgiPlusUsageCount;
      }
      else
         *vecptr++ = "";

      *vecptr++ = tkptr->TotalUsageCount;
      *vecptr++ = &tkptr->LastUsedBinaryTime;

      if (tkptr->RequestPtr != NULL)
      {
         *vecptr++ =
"<TR><TD></TD><TD ALIGN=left COLSPAN=1>!%%!AZ</TD>\
<TD ALIGN=left COLSPAN=6><TT>!HF!HF</TT></TD></TR>\n";

         if (tkptr->RequestPtr->RemoteUser[0])
         {
            *vecptr++ = "!HZ@";
            *vecptr++ = tkptr->RequestPtr->RemoteUser;
         }
         else
            *vecptr++ = "";
         *vecptr++ = tkptr->RequestPtr->rqNet.ClientHostName;
         *vecptr++ = tkptr->RequestPtr->ScriptName;
         *vecptr++ = tkptr->RequestPtr->rqHeader.PathInfoPtr;
      }
      else
         *vecptr++ = "";

      status = NetWriteFaol (rqptr, TaskFao, &FaoVector);
      if (VMSnok (status)) ErrorNoticed (status, "NetWriteFaol()", FI_LI);
   }

   /* script name cache entries */
   status = NetWriteFaol (rqptr, ScriptNameCacheFao, NULL);
   if (VMSnok (status)) ErrorNoticed (status, "NetWriteFaol()", FI_LI);

   Count = 0;
   for (leptr = DclScriptNameCacheList.HeadPtr;
        leptr != NULL;
        leptr = leptr->NextPtr)
   {
      captr = (struct DclScriptNameCacheStruct*)leptr;

      vecptr = FaoVector;
      *vecptr++ = ++Count;
      *vecptr++ = captr->ScriptName;
      *vecptr++ = captr->ResFileName;
      *vecptr++ = captr->HitCount;
      *vecptr++ = captr->UsageCount;
      *vecptr++ = &captr->LastBinaryTime;

      status = NetWriteFaol (rqptr, NameCacheFao, &FaoVector);
      if (VMSnok (status)) ErrorNoticed (status, "NetWriteFaol()", FI_LI);
   }

   vecptr = FaoVector;
   *vecptr++ = INTERNAL_CONTROL_DCL_PURGE;
   *vecptr++ = INTERNAL_CONTROL_DCL_DELETE;

   status = NetWriteFaol (rqptr, ButtonsFao, &FaoVector);
   if (VMSnok (status)) ErrorNoticed (status, "NetWriteFaol()", FI_LI);

   SysDclAst (NextTaskFunction, rqptr);
}

/*****************************************************************************/
/*
************
*** NOTE ***  This function takes a pointer to a request!!!
************  Due to it being a general control processing function.

Control function to delete and mark-for-delete all subprocesses. This is a
quick and effective means for purging all zombie and CGIplus subprocesses
forcing the next activation to reload the (possibly new) image.  The
'WithExtremePrejudice' option deletes subprocesses unconditionally, processing
or not.  It can be used to clear errant scripts without restarting the server. 
Called from Admin.c module.
*/

DclPurgeAllSubprocesses
(
struct RequestStruct *rqptr,
void *NextTaskFunction,
boolean WithExtremePrejudice
)
{
   register struct  ListEntryStruct  *leptr;
   register struct DclTaskStruct  *tkptr;

   int  status,
        DeletedCount,
        MarkedCount;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DclPurgeAllSubprocesses()\n");

   DclPurgeScriptNameCache ();

   DclPurgeAllSubprocessesCount++;

   DeletedCount = MarkedCount = 0;

   for (leptr = DclTaskList.HeadPtr; leptr != NULL; leptr = leptr->NextPtr)
   {
      tkptr = (struct DclTaskStruct*)leptr;

      if (!tkptr->SubprocessPid) continue;

      tkptr->MarkedForDelete = true;

      if (WithExtremePrejudice ||
          (tkptr->QueuedSysCommand <= tkptr->QueuedSysCommandAllowed &&
           !tkptr->QueuedSysOutput &&
           !tkptr->QueuedCgiPlusIn &&
           !tkptr->QueuedHttpInput &&
           !tkptr->FindScript &&
           tkptr->RequestPtr == NULL))
      {
         /* forced delete or subprocess not currently active, abort task */
         DclConcludeTask (tkptr, true);
         DeletedCount++;
      }
      else
      {
         /* subprocess is currently active, just leave marked for delete */
         MarkedCount++;
      }
   }

   WriteFaoStdout (
"%!AZ-I-ADMIN, !20%D, !AZ.\'!HZ\'@!AZ, purge DCL subprocesses\n\
!UL deleted, !UL marked for deletion\n",
      Utility, 0, rqptr->RemoteUser, rqptr->rqAuth.RealmDescrPtr,
      rqptr->rqNet.ClientHostName, DeletedCount, MarkedCount);

   if (OpcomMessages & OPCOM_ADMIN)
      WriteFaoOpcom (
"%!AZ-I-ADMIN, !AZ.\'!HZ\'@!AZ, purge DCL subprocesses\r\n\
!UL deleted, !UL marked for deletion",
         Utility, rqptr->RemoteUser, rqptr->rqAuth.RealmDescrPtr,
         rqptr->rqNet.ClientHostName, DeletedCount, MarkedCount);

   rqptr->rqResponse.PreExpired = PRE_EXPIRE_ADMIN;
   ReportSuccess (rqptr,
"Server !AZ had !UL process!AZ deleted, !UL marked for delete.",
      ServerHostPort, DeletedCount, DeletedCount == 1 ? "" : "es", MarkedCount);

   SysDclAst (NextTaskFunction, rqptr);
}

/*****************************************************************************/
/*
Same as for DclPurgeAllSubprocesses() except called from control module from
command line!
*/

char* DclControlPurgeAllSubprocesses (boolean WithExtremePrejudice)

{
   static char  Response [64];
   static $DESCRIPTOR (ResponseFaoDsc, "!UL deleted, !UL marked for delete");
   static $DESCRIPTOR (ResponseDsc, Response);

   register struct  ListEntryStruct  *leptr;
   register struct DclTaskStruct  *tkptr;

   int  status,
        DeletedCount,
        MarkedCount;
   unsigned short  Length;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DclControlPurgeAllSubprocesses()\n");

   DclPurgeScriptNameCache ();

   DclPurgeAllSubprocessesCount++;

   DeletedCount = MarkedCount = 0;

   for (leptr = DclTaskList.HeadPtr; leptr != NULL; leptr = leptr->NextPtr)
   {
      tkptr = (struct DclTaskStruct*)leptr;

      if (!tkptr->SubprocessPid) continue;

      tkptr->MarkedForDelete = true;

      if (WithExtremePrejudice ||
          (tkptr->QueuedSysCommand <= tkptr->QueuedSysCommandAllowed &&
           !tkptr->QueuedSysOutput &&
           !tkptr->QueuedCgiPlusIn &&
           !tkptr->QueuedHttpInput &&
           !tkptr->FindScript &&
           tkptr->RequestPtr == NULL))
      {
         /* forced delete or subprocess not currently active, abort task */
         DclConcludeTask (tkptr, true);
         DeletedCount++;
      }
      else
      {
         /* subprocess is currently active, just leave marked for delete */
         MarkedCount++;
      }
   }

   sys$fao (&ResponseFaoDsc, &Length, &ResponseDsc,
            DeletedCount, MarkedCount);
   Response[Length] = '\0';

   return (Response);
}

/*****************************************************************************/
/*
A mapping rule reload is a particularly "dangerous" time for scripting as
confusion could ensure about which CGIplus script is which and which script
name in the cache is which.  Hence at a mapping rule reload purge DCL suprocess
tasks and the script name cache.
*/

DclLoadedMappingRules ()

{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DclLoadedMappingRules()\n");

   DclPurgeScriptNameCache ();

   DclControlPurgeAllSubprocesses (false);
}

/*****************************************************************************/
/*
*/

DclTaskItemDebug
(
struct ListEntryStruct *leptr,
struct DclTaskStruct *tkptr
)
{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DclTaskItemDebug()\n");

#ifdef DBUG

   fprintf (stdout, "leptr %d %d %d (%d)\n",
            leptr->PrevPtr, leptr, leptr->NextPtr, tkptr);

   fprintf (stdout, "%08.08X [%d] %d %d %d %d %d %d %d |%s| %d\n",
            tkptr->SubprocessPid, tkptr->RequestPtr, tkptr->FindScript,
            tkptr->CgiPlusUsageCount, tkptr->TotalUsageCount,
            tkptr->QueuedSysCommand, tkptr->QueuedSysOutput,
            tkptr->QueuedCgiPlusIn, tkptr->QueuedHttpInput,
            tkptr->ScriptName, tkptr->LifeTimeCount);

#endif /* DBUG */

}

/*****************************************************************************/

